##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Make sure the supercritical steam cycle example solves.
"""

__author__ = "Miguel Zamarripa"

import pytest
from pyomo.environ import (
    SolverFactory, SolverStatus, TerminationCondition, value)
from pyomo.util.check_units import assert_units_consistent

import idaes.power_generation.flowsheets.\
    supercritical_power_plant.boiler_subflowsheet_build as blr
import idaes.power_generation.flowsheets.\
    supercritical_power_plant.SCPC_full_plant as SCPC
from idaes.core.util.model_statistics import (degrees_of_freedom)
from idaes.generic_models.properties import iapws95


@pytest.fixture(scope="module")
def boiler():
    m, solver = blr.main()

    m.solver = solver

    return m


@pytest.mark.integration
def test_init(boiler):
    # initialize each unit at the time
    blr.initialize(boiler)
    blr.unfix_inlets(boiler)
    # check that the model solved properly and has 0 degrees of freedom
    assert(degrees_of_freedom(boiler) == 0)


@pytest.mark.integration
def test_unit_consistency(boiler):
    assert_units_consistent(boiler)


@pytest.mark.integration
def test_boiler(boiler):
    # unfix inlets to build arcs at the flowsheet level
    boiler.fs.ATMP1.outlet.enth_mol[0].fix(62710.01)
    boiler.fs.ATMP1.SprayWater.flow_mol[0].unfix()
    result = boiler.solver.solve(boiler, tee=False)
    assert result.solver.termination_condition == \
        TerminationCondition.optimal
    assert result.solver.status == SolverStatus.ok
    assert value(boiler.fs.ECON.side_1.properties_out[0].temperature) == \
        pytest.approx(521.009, 1)


@pytest.mark.integration
def test_power_plant():
    # SCPC.main imports and solves the SCPC Power Plant Flowsheet
    m, results = SCPC.main()
    assert results.solver.status == SolverStatus.ok



import logging
import os
from pathlib import Path
from shutil import rmtree


def create_module_scratch(module_name):
    """Create a scratch that is easily found (by developers) and removed, and doesn't
    clutter up the working directories.
    """
    path = Path("~").expanduser()
    for path_part in ".idaes", "_scratch", module_name:
        path = path / path_part
        if not path.exists():
            path.mkdir()
    return path


def rmtree_scratch(scratch_path):
    """Do our level best to remove all the temporary files.
    """
    try:
        rmtree(scratch_path, ignore_errors=True)
    except Exception as err:
        pass



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
""" Tests for the controller model subclass of block
"""

import pytest
import pyomo.environ as pyo

from idaes.apps.caprese.tests.test_simple_model import (
        make_model,
        make_small_model,
        initialize_t0,
        copy_values_forward,
        )
from idaes.apps.caprese.controller import (
        _ControllerBlockData,
        ControllerBlock,
        SimpleControllerBlock,
        IndexedControllerBlock,
        )
from idaes.apps.caprese.nmpc_var import (
        DiffVar,
        AlgVar,
        InputVar,
        FixedVar,
        DerivVar,
        )
from idaes.apps.caprese.common.config import (
        ControlPenaltyType,
        )
from idaes.core.util.model_statistics import degrees_of_freedom

solver_available = pyo.SolverFactory('ipopt').available()
if solver_available:
    solver = pyo.SolverFactory('ipopt')
else:
    solver = None

class TestControllerBlock(object):

    @pytest.mark.unit
    def test_construct(self):
        model = make_small_model()
        time = model.time
        t0 = time.first()
        inputs = [model.flow_in[t0]]
        controller = ControllerBlock(
                model=model,
                time=time,
                inputs=inputs,
                )
        assert type(controller) is SimpleControllerBlock
        assert isinstance(controller, ControllerBlock)
        assert isinstance(controller, _ControllerBlockData)

        controller.construct()
        assert controller[None] is controller
        assert controller.mod is model
        assert controller.time is time
        assert all(i1 is i2 for i1, i2 in zip(controller._inputs, inputs))

    @pytest.mark.unit
    def test_construct_indexed(self):
        controller_set = pyo.Set(initialize=[0,1,2])
        controller_set.construct()
        horizon_map = {0: 1., 1: 3., 2: 5.}
        nfe_map = {0: 2, 1: 6, 2: 10}
        model_map = {i: make_model(horizon_map[i], nfe_map[i])
                for i in controller_set}
        time_map = {i: model_map[i].time for i in controller_set}
        inputs_map = {i: [model_map[i].flow_in[0]] for i in controller_set}

        controller = ControllerBlock(
                controller_set,
                model=model_map,
                time=time_map,
                inputs=inputs_map,
                )

        assert type(controller) is IndexedControllerBlock
        assert isinstance(controller, IndexedControllerBlock)
        controller.construct()
        assert all(b.parent_component() is controller for b in controller.values())

        for i in controller_set:
            assert i in controller
        for i, c in controller.items():
            assert c.mod is model_map[i]
            assert c.time is time_map[i]
            assert all(i1 is i2 for i1, i2 in zip(c._inputs, inputs_map[i]))

    @pytest.mark.unit
    def make_controller(self, sample_time=0.5, horizon=1., nfe=2):
        model = make_model(horizon=horizon, nfe=nfe)
        time = model.time
        t0 = time.first()
        inputs = [model.flow_in[t0]]
        controller = ControllerBlock(
                model=model,
                time=time,
                inputs=inputs,
                )
        controller.construct()
        controller.set_sample_time(sample_time)
        return controller

    @pytest.mark.unit
    def test_add_setpoint_objective(self):
        controller = self.make_controller()
        time = controller.time
        t0 = time.first()
        setpoint = [
                (controller.mod.flow_in[t0], 3.0),
                ]
        weights = [
                (controller.mod.flow_in[t0], 2.0),
                ]
        controller.mod.flow_in[:].set_value(3.0)
        initialize_t0(controller.mod)

        controller.add_setpoint_objective(setpoint, weights)

        assert hasattr(controller, 'setpoint_objective')

        pred_obj_expr = (2.0*(controller.mod.flow_in[t0] - 3.0)**2)
        obj_expr = controller.setpoint_objective.expr
        assert pyo.value(pred_obj_expr) == pyo.value(obj_expr)
        assert pred_obj_expr.to_string() == obj_expr.to_string()

        controller.del_component(controller.setpoint_objective)

        setpoint = [
                (controller.mod.flow_in[t0], 3.0),
                (controller.mod.conc[t0,'A'], 1.5),
                ]
        weights = [
                (controller.mod.flow_in[t0], 1.0),
                (controller.mod.conc[t0,'A'], 5.0),
                ]
        controller.add_setpoint_objective(setpoint, weights)
        pred_obj_expr = (
                1.0*(controller.mod.flow_in[t0] - 3.0)**2 + 
                5.0*(controller.mod.conc[t0,'A'] - 1.5)**2
                )
        obj_expr = controller.setpoint_objective.expr
        assert pyo.value(pred_obj_expr) == pyo.value(obj_expr)
        assert pred_obj_expr.to_string() == obj_expr.to_string()

    @pytest.mark.component
    @pytest.mark.skipif(not solver_available, reason='IPOPT is not available')
    def test_solve_setpoint_steady(self):
        controller = self.make_controller()
        time = controller.time
        t0 = time.first()
        setpoint = [
                (controller.mod.flow_in[t0], 3.0),
                ]
        weights = [
                (controller.mod.flow_in[t0], 1.0),
                ]
        controller.add_setpoint_objective(setpoint, weights)
        controller.mod.flow_in[:].set_value(3.0)
        initialize_t0(controller.mod)

        dof_prior = degrees_of_freedom(controller)
        controller.solve_setpoint(solver, require_steady=True)
        dof_post = degrees_of_freedom(controller)

        assert dof_prior == dof_post

        assert controller.differential_vars[0].setpoint == \
                pytest.approx(3.75, abs=1e-3)
        assert controller.differential_vars[1].setpoint == \
                pytest.approx(1.25, abs=1e-3)
        assert controller.algebraic_vars[0].setpoint == \
                pytest.approx(3.0, abs=1e-3)
        assert controller.algebraic_vars[1].setpoint == \
                pytest.approx(-3.75, abs=1e-3)
        assert controller.algebraic_vars[2].setpoint == \
                pytest.approx(3.75, abs=1e-3)
        assert controller.input_vars[0].setpoint == \
                pytest.approx(3.0, abs=1e-3)

    @pytest.mark.component
    @pytest.mark.skipif(not solver_available, reason='IPOPT is not available')
    def test_solve_setpoint_unsteady(self):
        controller = self.make_controller()
        time = controller.time
        t0 = time.first()
        setpoint = [
                (controller.mod.flow_in[t0], 3.0),
                (controller.mod.conc[t0,'A'], 5.0),
                ]
        weights = [
                (controller.mod.flow_in[t0], 1.0),
                (controller.mod.conc[t0,'A'], 1.0),
                ]
        controller.add_setpoint_objective(setpoint, weights)
        controller.mod.flow_in[:].set_value(3.0)
        initialize_t0(controller.mod)

        dof_prior = degrees_of_freedom(controller)
        controller.solve_setpoint(solver, require_steady=False)
        dof_post = degrees_of_freedom(controller)

        assert dof_prior == dof_post

        assert controller.differential_vars[0].setpoint == \
                pytest.approx(5.0, abs=1e-3)
        assert controller.differential_vars[1].setpoint == \
                pytest.approx(7.11, abs=1e-3)
        assert controller.algebraic_vars[0].setpoint == \
                pytest.approx(3.0, abs=1e-3)
        assert controller.algebraic_vars[1].setpoint == \
                pytest.approx(-5.0, abs=1e-3)
        assert controller.algebraic_vars[2].setpoint == \
                pytest.approx(5.0, abs=1e-3)
        assert controller.input_vars[0].setpoint == \
                pytest.approx(3.0, abs=1e-3)

    @pytest.mark.component
    @pytest.mark.skipif(not solver_available, reason='IPOPT is not available')
    def test_add_tracking_objective(self):
        controller = self.make_controller()
        time = controller.time
        t0 = time.first()
        setpoint = [
                (controller.mod.flow_in[t0], 3.0),
                ]
        weights = [
                (controller.mod.flow_in[t0], 1.0),
                ]
        controller.mod.flow_in[:].set_value(3.0)
        initialize_t0(controller.mod)
        copy_values_forward(controller.mod)
        controller.add_setpoint_objective(setpoint, weights)
        controller.solve_setpoint(solver)

        # Re-initialize inputs so they are not at the setpoint
        # for objective evaluation.
        controller.mod.flow_in[:].set_value(2.5)
        
        weights = [
                (controller.mod.conc[t0,'A'], 1),
                (controller.mod.conc[t0,'B'], 1),
                (controller.mod.rate[t0,'A'], 1),
                (controller.mod.rate[t0,'B'], 1),
                (controller.mod.flow_out[t0], 1),
                (controller.mod.flow_in[t0], 1),
                ]

        # Construct predicted objective functions:
        pred_obj = {i: 0. for i in range(1, 4)}
        sample_points = controller.sample_points[1:]
        sample_time = controller.sample_time
        for v in controller.component_objects(DiffVar):
            for t in sample_points:
                pred_obj[1] += (v[t] - v.setpoint)**2
                pred_obj[2] += (v[t] - v.setpoint)**2
                pred_obj[3] += (v[t] - v.setpoint)**2
        for v in controller.component_objects(AlgVar):
            for t in sample_points:
                pred_obj[3] += (v[t] - v.setpoint)**2
        for v in controller.component_objects(InputVar):
            for t in sample_points:
                i_prev = time.find_nearest_index(t-sample_time, tolerance=1e-8)
                t_prev = time[i_prev]
                pred_obj[1] += (v[t] - v.setpoint)**2
                pred_obj[2] += (v[t] - v[t_prev])**2
                pred_obj[3] += (v[t] - v[t_prev])**2

        controller.add_tracking_objective(
                weights,
                control_penalty_type=ControlPenaltyType.ERROR,
                state_ctypes=DiffVar,
                )
        assert pyo.value(controller.tracking_objective.expr == pred_obj[1])
        assert pyo.value(controller.tracking_objective.expr) > 0
        controller.del_component(controller.tracking_objective)

        controller.add_tracking_objective(
                weights,
                control_penalty_type=ControlPenaltyType.ACTION,
                state_ctypes=DiffVar,
                )
        assert pyo.value(controller.tracking_objective.expr == pred_obj[2])
        assert pyo.value(controller.tracking_objective.expr) > 0
        controller.del_component(controller.tracking_objective)

        controller.add_tracking_objective(
                weights,
                control_penalty_type=ControlPenaltyType.ACTION,
                state_ctypes=(DiffVar, AlgVar),
                )
        assert pyo.value(controller.tracking_objective.expr == pred_obj[3])
        assert pyo.value(controller.tracking_objective.expr) > 0

    @pytest.mark.unit
    def test_constrain_control_inputs_piecewise_constant(self):
        controller = self.make_controller()
        time = controller.time
        t0 = time.first()

        controller.constrain_control_inputs_piecewise_constant()
        assert hasattr(controller, 'pwc_constraint')
        sample_points = controller.sample_points
        sample_point_set = set(sample_points)

        for i, t in controller.INPUT_SET*controller.time:
            if t in sample_point_set:
                assert (i, t) not in controller.pwc_constraint
            else:
                assert (i, t) in controller.pwc_constraint
        for i, t in controller.pwc_constraint:
            pwc_expr = controller.pwc_constraint[i,t].expr
            tn = time.next(t)
            inputs = controller.vectors.input
            pred_expr = (inputs[i, tn] - inputs[i, t] == 0.)
            assert pwc_expr.to_string() == pred_expr.to_string()



from idaes.power_generation.unit_models.helm.splitter import HelmSplitter
from idaes.power_generation.unit_models.helm.valve_steam import HelmValve, ValveFunctionType
from idaes.power_generation.unit_models.helm.mixer import HelmMixer
from idaes.generic_models.unit_models import MomentumMixingType
from idaes.power_generation.unit_models.helm.compressor import HelmIsentropicCompressor
from idaes.power_generation.unit_models.helm.turbine import HelmIsentropicTurbine
from idaes.power_generation.unit_models.helm.pump import HelmPump
from idaes.power_generation.unit_models.helm.turbine_inlet import HelmTurbineInletStage
from idaes.power_generation.unit_models.helm.turbine_stage import HelmTurbineStage
from idaes.power_generation.unit_models.helm.turbine_outlet import HelmTurbineOutletStage
from idaes.power_generation.unit_models.helm.turbine_multistage import HelmTurbineMultistage
from idaes.power_generation.unit_models.helm.condenser_ntu import HelmNtuCondenser, HelmNtuCondenserData



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for FTP state formulation

Authors: Andrew Lee
"""

import pytest
from sys import modules

from pyomo.environ import (
    ConcreteModel, Constraint, Expression, Var, units as pyunits)
from pyomo.util.check_units import (
    check_units_equivalent, assert_units_consistent)

# Need define_default_scaling_factors, even though it is not used directly
from idaes.generic_models.properties.core.state_definitions.FTPx import \
    FTPx, define_state, set_metadata, define_default_scaling_factors
from idaes.core import (MaterialFlowBasis,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        declare_process_block_class)
from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterData)
from idaes.generic_models.properties.core.generic.tests.dummy_eos import DummyEoS
from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog


@declare_process_block_class("DummyParameterBlock")
class DummyParameterData(GenericParameterData):
    pass


@pytest.mark.unit
def test_set_metadata():
    assert set_metadata(None) is None


class TestInvalidBounds(object):

    @pytest.mark.unit
    def test_bad_name(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(default={
                "components": {"c1": {}, "c2": {}, "c3": {}},
                "phases": {
                    "p1": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K},
                "state_bounds": {"foo": (None, None, None)}})

        with pytest.raises(
                ConfigurationError,
                match="props\[1\] - found unexpected state_bounds key foo. "
                "Please ensure bounds are provided only for expected state "
                "variables and that you have typed the variable names "
                "correctly."):
            m.props = m.params.build_state_block(
                [1], default={"defined_state": True})

    @pytest.mark.unit
    def test_mole_frac(self, caplog):
        m = ConcreteModel()

        caplog.set_level(
            idaeslog.WARNING,
            logger=("idaes.generic_models.properties.core."))

        m.params = DummyParameterBlock(default={
                "components": {"c1": {}, "c2": {}, "c3": {}},
                "phases": {
                    "p1": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K},
                "state_bounds": {"mole_frac_comp": (None, None, None)}})

        # Create a dummy state block
        m.props = m.params.build_state_block(
            [1], default={"defined_state": True})

        assert ("props[1] - found state_bounds argument for mole_frac_comp."
                " Mole fraction bounds are set automatically and "
                "this argument will be ignored." in caplog.text)


class Test1PhaseDefinedStateFalseNoBounds(object):
    # Test define_state method with no bounds and defined_State = False
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(default={
                "components": {"c1": {}, "c2": {}, "c3": {}},
                "phases": {
                    "p1": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K}})

        # Create state block
        m.props = m.params.build_state_block(
            [1], default={"defined_state": False})

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_always_flash(self, frame):
        define_state(frame.props[1])

        assert frame.props[1].always_flash

    @pytest.mark.unit
    def test_vars(self, frame):
        # Check that all necessary variables have been constructed and have
        # the correct values
        assert isinstance(frame.props[1].flow_mol, Var)
        assert frame.props[1].flow_mol.value is None
        assert check_units_equivalent(frame.props[1].flow_mol,
                                      pyunits.mol/pyunits.s)

        assert isinstance(frame.props[1].mole_frac_comp, Var)
        assert len(frame.props[1].mole_frac_comp) == 3
        for i in frame.props[1].mole_frac_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].mole_frac_comp[i].value == 1/3
        assert check_units_equivalent(frame.props[1].mole_frac_comp,
                                      None)

        assert isinstance(frame.props[1].pressure, Var)
        assert frame.props[1].pressure.value is None
        assert check_units_equivalent(frame.props[1].pressure,
                                      pyunits.Pa)

        assert isinstance(frame.props[1].temperature, Var)
        assert frame.props[1].temperature.value is None
        assert check_units_equivalent(frame.props[1].temperature,
                                      pyunits.K)

        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 1
        for i in frame.props[1].flow_mol_phase:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].flow_mol_phase[i].value is None
        assert check_units_equivalent(frame.props[1].flow_mol_phase,
                                      pyunits.mol/pyunits.s)

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 1
        for i in frame.props[1].phase_frac:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].phase_frac[i].value == 1
        assert check_units_equivalent(frame.props[1].phase_frac,
                                      None)

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 3
        for i in frame.props[1].mole_frac_phase_comp:
            assert i in [("p1", "c1"), ("p1", "c2"), ("p1", "c3")]
            assert frame.props[1].mole_frac_phase_comp[i].value == 1/3
        assert check_units_equivalent(frame.props[1].mole_frac_phase_comp,
                                      None)

    @pytest.mark.unit
    def test_constraints(self, frame):
        # Check that the correct constraints are present
        assert isinstance(frame.props[1].total_flow_balance, Constraint)
        assert len(frame.props[1].total_flow_balance) == 1
        assert str(frame.props[1].total_flow_balance.body) == str(
            frame.props[1].flow_mol -
            frame.props[1].flow_mol_phase[frame.params.phase_list[1]])

        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3
        for i in frame.props[1].component_flow_balances:
            assert i in frame.props[1].params.component_list
            assert str(frame.props[1].component_flow_balances[i].body) == str(
                frame.props[1].mole_frac_comp[i] -
                frame.props[1].mole_frac_phase_comp[
                        frame.params.phase_list[1], i])

        assert isinstance(frame.props[1].sum_mole_frac_out, Constraint)
        assert len(frame.props[1].sum_mole_frac_out) == 1
        assert str(frame.props[1].sum_mole_frac_out.body) == str(
                sum(frame.props[1].mole_frac_comp[i]
                    for i in frame.props[1].params.component_list))

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 1
        for i in frame.props[1].phase_fraction_constraint:
            assert i in frame.props[1].params.phase_list
            assert str(frame.props[1].phase_fraction_constraint[i].body) == \
                str(frame.props[1].phase_frac[i])

        assert_units_consistent(frame.props[1])


class Test1PhaseDefinedStateTrueWithBounds(object):
    # Test define_state method with no bounds and defined_State = False
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        # Create a dummy parameter block
        m.params = DummyParameterBlock(default={
                "components": {"c1": {}, "c2": {}, "c3": {}},
                "phases": {
                    "p1": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "state_bounds": {"flow_mol": (0, 100, 200),
                                 "temperature": (290, 345,  400),
                                 "pressure": (1e5, 3e5, 5e5)},
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K}})

        # Create state block
        m.props = m.params.build_state_block(
            [1], default={"defined_state": True})

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_always_flash(self, frame):
        define_state(frame.props[1])

        assert frame.props[1].always_flash

    @pytest.mark.unit
    def test_vars(self, frame):
        # Check that all necessary variables have been constructed and have
        # the correct values
        assert isinstance(frame.props[1].flow_mol, Var)
        assert frame.props[1].flow_mol.value == 100
        assert frame.props[1].flow_mol.lb == 0
        assert frame.props[1].flow_mol.ub == 200
        assert check_units_equivalent(frame.props[1].flow_mol,
                                      pyunits.mol/pyunits.s)

        assert isinstance(frame.props[1].mole_frac_comp, Var)
        assert len(frame.props[1].mole_frac_comp) == 3
        for i in frame.props[1].mole_frac_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].mole_frac_comp[i].value == 1/3
        assert check_units_equivalent(frame.props[1].mole_frac_comp,
                                      None)

        assert isinstance(frame.props[1].pressure, Var)
        assert frame.props[1].pressure.value == 3e5
        assert frame.props[1].pressure.lb == 1e5
        assert frame.props[1].pressure.ub == 5e5
        assert check_units_equivalent(frame.props[1].pressure,
                                      pyunits.Pa)

        assert isinstance(frame.props[1].temperature, Var)
        assert frame.props[1].temperature.value == 345
        assert frame.props[1].temperature.lb == 290
        assert frame.props[1].temperature.ub == 400
        assert check_units_equivalent(frame.props[1].temperature,
                                      pyunits.K)

        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 1
        for i in frame.props[1].flow_mol_phase:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].flow_mol_phase[i].value == 100
            assert frame.props[1].flow_mol_phase[i].lb == 0
            assert frame.props[1].flow_mol_phase[i].ub == 200
        assert check_units_equivalent(frame.props[1].flow_mol_phase,
                                      pyunits.mol/pyunits.s)

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 1
        for i in frame.props[1].phase_frac:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].phase_frac[i].value == 1
        assert check_units_equivalent(frame.props[1].phase_frac,
                                      None)

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 3
        for i in frame.props[1].mole_frac_phase_comp:
            assert i in [("p1", "c1"), ("p1", "c2"), ("p1", "c3")]
            assert frame.props[1].mole_frac_phase_comp[i].value == 1/3
        assert check_units_equivalent(frame.props[1].mole_frac_phase_comp,
                                      None)

    @pytest.mark.unit
    def test_constraints(self, frame):
        # Check that the correct constraints are present
        assert isinstance(frame.props[1].total_flow_balance, Constraint)
        assert len(frame.props[1].total_flow_balance) == 1
        assert str(frame.props[1].total_flow_balance.body) == str(
            frame.props[1].flow_mol -
            frame.props[1].flow_mol_phase[frame.params.phase_list[1]])

        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3
        for i in frame.props[1].component_flow_balances:
            assert i in frame.props[1].params.component_list
            assert str(frame.props[1].component_flow_balances[i].body) == str(
                frame.props[1].mole_frac_comp[i] -
                frame.props[1].mole_frac_phase_comp[
                    frame.params.phase_list[1], i])

        assert not hasattr(frame.props[1], "sum_mole_frac_out")

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 1
        for i in frame.props[1].phase_fraction_constraint:
            assert i in frame.props[1].params.phase_list
            assert str(frame.props[1].phase_fraction_constraint[i].body) == \
                str(frame.props[1].phase_frac[i])

        assert_units_consistent(frame.props[1])


class Test2PhaseDefinedStateFalseNoBounds(object):
    # Test define_state method with no bounds and defined_State = False
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        # Create a dummy parameter block
        m.params = DummyParameterBlock(default={
                "components": {"c1": {}, "c2": {}, "c3": {}},
                "phases": {
                    "p1": {"equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K}})

        # Create state block
        m.props = m.params.build_state_block(
            [1], default={"defined_state": False})

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_always_flash(self, frame):
        define_state(frame.props[1])

        assert frame.props[1].always_flash

    @pytest.mark.unit
    def test_vars(self, frame):
        # Check that all necessary variables have been constructed and have
        # the correct values
        assert isinstance(frame.props[1].flow_mol, Var)
        assert frame.props[1].flow_mol.value is None
        assert check_units_equivalent(frame.props[1].flow_mol,
                                      pyunits.mol/pyunits.s)

        assert isinstance(frame.props[1].mole_frac_comp, Var)
        assert len(frame.props[1].mole_frac_comp) == 3
        for i in frame.props[1].mole_frac_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].mole_frac_comp[i].value == 1/3
        assert check_units_equivalent(frame.props[1].mole_frac_comp,
                                      None)

        assert isinstance(frame.props[1].pressure, Var)
        assert frame.props[1].pressure.value is None
        assert check_units_equivalent(frame.props[1].pressure,
                                      pyunits.Pa)

        assert isinstance(frame.props[1].temperature, Var)
        assert frame.props[1].temperature.value is None
        assert check_units_equivalent(frame.props[1].temperature,
                                      pyunits.K)

        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 2
        for i in frame.props[1].flow_mol_phase:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].flow_mol_phase[i].value is None
        assert check_units_equivalent(frame.props[1].flow_mol_phase,
                                      pyunits.mol/pyunits.s)

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 2
        for i in frame.props[1].phase_frac:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].phase_frac[i].value == 1/2
        assert check_units_equivalent(frame.props[1].phase_frac,
                                      None)

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 6
        for i in frame.props[1].mole_frac_phase_comp:
            assert i in [("p1", "c1"), ("p1", "c2"), ("p1", "c3"),
                         ("p2", "c1"), ("p2", "c2"), ("p2", "c3")]
            assert frame.props[1].mole_frac_phase_comp[i].value == 1/3
        assert check_units_equivalent(frame.props[1].mole_frac_phase_comp,
                                      None)

    @pytest.mark.unit
    def test_constraints(self, frame):
        # Check that the correct constraints are present
        assert isinstance(frame.props[1].total_flow_balance, Constraint)
        assert len(frame.props[1].total_flow_balance) == 1
        assert str(frame.props[1].total_flow_balance.body) == str(
                sum(frame.props[1].flow_mol_phase[p]
                    for p in frame.props[1].params.phase_list) -
                frame.props[1].flow_mol)

        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3
        for i in frame.props[1].component_flow_balances:
            assert i in frame.props[1].params.component_list
            assert str(frame.props[1].component_flow_balances[i].body) == str(
                frame.props[1].flow_mol * frame.props[1].mole_frac_comp[i] -
                sum(frame.props[1].flow_mol_phase[p] *
                    frame.props[1].mole_frac_phase_comp[p, i]
                    for p in frame.props[1].params.phase_list))

        assert isinstance(frame.props[1].sum_mole_frac, Constraint)
        assert len(frame.props[1].sum_mole_frac) == 1
        assert str(frame.props[1].sum_mole_frac.body) == str(
                sum(frame.props[1].mole_frac_phase_comp[
                    frame.props[1].params.phase_list[1], i]
                    for i in frame.props[1].params.component_list) -
                sum(frame.props[1].mole_frac_phase_comp[
                    frame.props[1].params.phase_list[2], i]
                    for i in frame.props[1].params.component_list))

        assert isinstance(frame.props[1].sum_mole_frac_out, Constraint)
        assert len(frame.props[1].sum_mole_frac_out) == 1
        assert str(frame.props[1].sum_mole_frac_out.body) == str(
                sum(frame.props[1].mole_frac_comp[i]
                    for i in frame.props[1].params.component_list))

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 2
        for i in frame.props[1].phase_fraction_constraint:
            assert i in frame.props[1].params.phase_list
            assert str(frame.props[1].phase_fraction_constraint[i].body) == \
                str(frame.props[1].phase_frac[i]*frame.props[1].flow_mol -
                    frame.props[1].flow_mol_phase[i])

        assert_units_consistent(frame.props[1])


class Test2PhaseDefinedStateTrueWithBounds(object):
    # Test define_state method with no bounds and defined_State = False
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        # Create a dummy parameter block
        m.params = DummyParameterBlock(default={
                "components": {"c1": {}, "c2": {}, "c3": {}},
                "phases": {
                    "p1": {"equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "state_bounds": {"flow_mol": (0, 100, 200),
                                 "temperature": (290, 345, 400),
                                 "pressure": (1e5, 3e5, 5e5)},
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K}})

        # Create state block
        m.props = m.params.build_state_block(
            [1], default={"defined_state": True})

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_always_flash(self, frame):
        define_state(frame.props[1])

        assert frame.props[1].always_flash

    @pytest.mark.unit
    def test_vars(self, frame):
        # Check that all necessary variables have been constructed and have
        # the correct values
        assert isinstance(frame.props[1].flow_mol, Var)
        assert frame.props[1].flow_mol.value == 100
        assert frame.props[1].flow_mol.lb == 0
        assert frame.props[1].flow_mol.ub == 200
        assert check_units_equivalent(frame.props[1].flow_mol,
                                      pyunits.mol/pyunits.s)

        assert isinstance(frame.props[1].mole_frac_comp, Var)
        assert len(frame.props[1].mole_frac_comp) == 3
        for i in frame.props[1].mole_frac_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].mole_frac_comp[i].value == 1/3
        assert check_units_equivalent(frame.props[1].mole_frac_comp,
                                      None)

        assert isinstance(frame.props[1].pressure, Var)
        assert frame.props[1].pressure.value == 3e5
        assert frame.props[1].pressure.lb == 1e5
        assert frame.props[1].pressure.ub == 5e5
        assert check_units_equivalent(frame.props[1].pressure,
                                      pyunits.Pa)

        assert isinstance(frame.props[1].temperature, Var)
        assert frame.props[1].temperature.value == 345
        assert frame.props[1].temperature.lb == 290
        assert frame.props[1].temperature.ub == 400
        assert check_units_equivalent(frame.props[1].temperature,
                                      pyunits.K)

        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 2
        for i in frame.props[1].flow_mol_phase:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].flow_mol_phase[i].value == 100/2
            assert frame.props[1].flow_mol_phase[i].lb == 0
            assert frame.props[1].flow_mol_phase[i].ub == 200
        assert check_units_equivalent(frame.props[1].flow_mol_phase,
                                      pyunits.mol/pyunits.s)

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 2
        for i in frame.props[1].phase_frac:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].phase_frac[i].value == 1/2
        assert check_units_equivalent(frame.props[1].phase_frac,
                                      None)

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 6
        for i in frame.props[1].mole_frac_phase_comp:
            assert i in [("p1", "c1"), ("p1", "c2"), ("p1", "c3"),
                         ("p2", "c1"), ("p2", "c2"), ("p2", "c3")]
            assert frame.props[1].mole_frac_phase_comp[i].value == 1/3
        assert check_units_equivalent(frame.props[1].mole_frac_phase_comp,
                                      None)

    @pytest.mark.unit
    def test_constraints(self, frame):
        # Check that the correct constraints are present
        assert isinstance(frame.props[1].total_flow_balance, Constraint)
        assert len(frame.props[1].total_flow_balance) == 1
        assert str(frame.props[1].total_flow_balance.body) == str(
                sum(frame.props[1].flow_mol_phase[p]
                    for p in frame.props[1].params.phase_list) -
                frame.props[1].flow_mol)

        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3
        for i in frame.props[1].component_flow_balances:
            assert i in frame.props[1].params.component_list
            assert str(frame.props[1].component_flow_balances[i].body) == str(
                frame.props[1].flow_mol * frame.props[1].mole_frac_comp[i] -
                sum(frame.props[1].flow_mol_phase[p] *
                    frame.props[1].mole_frac_phase_comp[p, i]
                    for p in frame.props[1].params.phase_list))

        assert isinstance(frame.props[1].sum_mole_frac, Constraint)
        assert len(frame.props[1].sum_mole_frac) == 1
        assert str(frame.props[1].sum_mole_frac.body) == str(
                sum(frame.props[1].mole_frac_phase_comp[
                    frame.props[1].params.phase_list[1], i]
                    for i in frame.props[1].params.component_list) -
                sum(frame.props[1].mole_frac_phase_comp[
                    frame.props[1].params.phase_list[2], i]
                    for i in frame.props[1].params.component_list))

        assert not hasattr(frame.props[1], "sum_mole_frac_out")

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 2
        for i in frame.props[1].phase_fraction_constraint:
            assert i in frame.props[1].params.phase_list
            assert str(frame.props[1].phase_fraction_constraint[i].body) == \
                str(frame.props[1].phase_frac[i]*frame.props[1].flow_mol -
                    frame.props[1].flow_mol_phase[i])

        assert_units_consistent(frame.props[1])


class Test3PhaseDefinedStateFalseNoBounds(object):
    # Test define_state method with no bounds and defined_State = False
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        # Create a dummy parameter block
        m.params = DummyParameterBlock(default={
                "components": {"c1": {}, "c2": {}, "c3": {}},
                "phases": {
                    "p1": {"equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS},
                    "p3": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K}})

        # Create state block
        m.props = m.params.build_state_block(
            [1], default={"defined_state": False})

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_always_flash(self, frame):
        define_state(frame.props[1])

        assert frame.props[1].always_flash

    @pytest.mark.unit
    def test_vars(self, frame):
        # Check that all necessary variables have been constructed and have
        # the correct values
        assert isinstance(frame.props[1].flow_mol, Var)
        assert frame.props[1].flow_mol.value is None
        assert check_units_equivalent(frame.props[1].flow_mol,
                                      pyunits.mol/pyunits.s)

        assert isinstance(frame.props[1].mole_frac_comp, Var)
        assert len(frame.props[1].mole_frac_comp) == 3
        for i in frame.props[1].mole_frac_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].mole_frac_comp[i].value == 1/3
        assert check_units_equivalent(frame.props[1].mole_frac_comp,
                                      None)

        assert isinstance(frame.props[1].pressure, Var)
        assert frame.props[1].pressure.value is None
        assert check_units_equivalent(frame.props[1].pressure,
                                      pyunits.Pa)

        assert isinstance(frame.props[1].temperature, Var)
        assert frame.props[1].temperature.value is None
        assert check_units_equivalent(frame.props[1].temperature,
                                      pyunits.K)

        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 3
        for i in frame.props[1].flow_mol_phase:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].flow_mol_phase[i].value is None
        assert check_units_equivalent(frame.props[1].flow_mol_phase,
                                      pyunits.mol/pyunits.s)

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 3
        for i in frame.props[1].phase_frac:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].phase_frac[i].value == 1/3
        assert check_units_equivalent(frame.props[1].phase_frac,
                                      None)

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 9
        for i in frame.props[1].mole_frac_phase_comp:
            assert i in [("p1", "c1"), ("p1", "c2"), ("p1", "c3"),
                         ("p2", "c1"), ("p2", "c2"), ("p2", "c3"),
                         ("p3", "c1"), ("p3", "c2"), ("p3", "c3")]
            assert frame.props[1].mole_frac_phase_comp[i].value == 1/3
        assert check_units_equivalent(frame.props[1].mole_frac_phase_comp,
                                      None)

    @pytest.mark.unit
    def test_constraints(self, frame):
        # Check that the correct constraints are present
        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3
        for j in frame.props[1].component_flow_balances:
            assert j in frame.params.component_list
            assert str(frame.props[1].component_flow_balances[j].body) == str(
                frame.props[1].flow_mol*frame.props[1].mole_frac_comp[j] -
                sum(frame.props[1].flow_mol_phase[p] *
                    frame.props[1].mole_frac_phase_comp[p, j]
                    for p in frame.props[1].params.phase_list))

        assert isinstance(frame.props[1].sum_mole_frac, Constraint)
        assert len(frame.props[1].sum_mole_frac) == 3
        for p in frame.props[1].sum_mole_frac:
            assert p in frame.params.phase_list
            assert str(frame.props[1].sum_mole_frac[p].body) == str(
                    sum(frame.props[1].mole_frac_phase_comp[p, i]
                        for i in frame.props[1].params.component_list))

        assert isinstance(frame.props[1].sum_mole_frac_out, Constraint)
        assert len(frame.props[1].sum_mole_frac_out) == 1
        assert str(frame.props[1].sum_mole_frac_out.body) == str(
                sum(frame.props[1].mole_frac_comp[i]
                    for i in frame.props[1].params.component_list))

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 3
        for i in frame.props[1].phase_fraction_constraint:
            assert i in frame.props[1].params.phase_list
            assert str(frame.props[1].phase_fraction_constraint[i].body) == \
                str(frame.props[1].phase_frac[i]*frame.props[1].flow_mol -
                    frame.props[1].flow_mol_phase[i])

        assert_units_consistent(frame.props[1])


class Test3PhaseDefinedStateTrueWithBounds(object):
    # Test define_state method with no bounds and defined_State = False
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        # Create a dummy parameter block
        m.params = DummyParameterBlock(default={
                "components": {"c1": {}, "c2": {}, "c3": {}},
                "phases": {
                    "p1": {"equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS},
                    "p3": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "state_bounds": {"flow_mol": (0, 100, 200),
                                 "temperature": (290, 345, 400),
                                 "pressure": (1e5, 3e5, 5e5)},
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K}})

        # Create state block
        m.props = m.params.build_state_block(
            [1], default={"defined_state": True})

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_always_flash(self, frame):
        define_state(frame.props[1])

        assert frame.props[1].always_flash

    @pytest.mark.unit
    def test_vars(self, frame):
        # Check that all necessary variables have been constructed and have
        # the correct values
        assert isinstance(frame.props[1].flow_mol, Var)
        assert frame.props[1].flow_mol.value == 100
        assert frame.props[1].flow_mol.lb == 0
        assert frame.props[1].flow_mol.ub == 200
        assert check_units_equivalent(frame.props[1].flow_mol,
                                      pyunits.mol/pyunits.s)

        assert isinstance(frame.props[1].mole_frac_comp, Var)
        assert len(frame.props[1].mole_frac_comp) == 3
        for i in frame.props[1].mole_frac_comp:
            assert i in frame.props[1].params.component_list
            assert frame.props[1].mole_frac_comp[i].value == 1/3
        assert check_units_equivalent(frame.props[1].mole_frac_comp,
                                      None)

        assert isinstance(frame.props[1].pressure, Var)
        assert frame.props[1].pressure.value == 3e5
        assert frame.props[1].pressure.lb == 1e5
        assert frame.props[1].pressure.ub == 5e5
        assert check_units_equivalent(frame.props[1].pressure,
                                      pyunits.Pa)

        assert isinstance(frame.props[1].temperature, Var)
        assert frame.props[1].temperature.value == 345
        assert frame.props[1].temperature.lb == 290
        assert frame.props[1].temperature.ub == 400
        assert check_units_equivalent(frame.props[1].temperature,
                                      pyunits.K)

        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 3
        for i in frame.props[1].flow_mol_phase:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].flow_mol_phase[i].value == 100/3
            assert frame.props[1].flow_mol_phase[i].lb == 0
            assert frame.props[1].flow_mol_phase[i].ub == 200
        assert check_units_equivalent(frame.props[1].flow_mol_phase,
                                      pyunits.mol/pyunits.s)

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 3
        for i in frame.props[1].phase_frac:
            assert i in frame.props[1].params.phase_list
            assert frame.props[1].phase_frac[i].value == 1/3
        assert check_units_equivalent(frame.props[1].phase_frac,
                                      None)

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 9
        for i in frame.props[1].mole_frac_phase_comp:
            assert i in [("p1", "c1"), ("p1", "c2"), ("p1", "c3"),
                         ("p2", "c1"), ("p2", "c2"), ("p2", "c3"),
                         ("p3", "c1"), ("p3", "c2"), ("p3", "c3")]
            assert frame.props[1].mole_frac_phase_comp[i].value == 1/3
        assert check_units_equivalent(frame.props[1].mole_frac_phase_comp,
                                      None)

    @pytest.mark.unit
    def test_constraints(self, frame):
        # Check that the correct constraints are present
        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3
        for j in frame.props[1].component_flow_balances:
            assert j in frame.params.component_list
            assert str(frame.props[1].component_flow_balances[j].body) == str(
                frame.props[1].flow_mol*frame.props[1].mole_frac_comp[j] -
                sum(frame.props[1].flow_mol_phase[p] *
                    frame.props[1].mole_frac_phase_comp[p, j]
                    for p in frame.props[1].params.phase_list))

        assert isinstance(frame.props[1].sum_mole_frac, Constraint)
        assert len(frame.props[1].sum_mole_frac) == 3
        for p in frame.props[1].sum_mole_frac:
            assert p in frame.params.phase_list
            assert str(frame.props[1].sum_mole_frac[p].body) == str(
                    sum(frame.props[1].mole_frac_phase_comp[p, i]
                        for i in frame.props[1].params.component_list))

        assert not hasattr(frame.props[1], "sum_mole_frac_out")

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 3
        for i in frame.props[1].phase_fraction_constraint:
            assert i in frame.props[1].params.phase_list
            assert str(frame.props[1].phase_fraction_constraint[i].body) == \
                str(frame.props[1].phase_frac[i]*frame.props[1].flow_mol -
                    frame.props[1].flow_mol_phase[i])

        assert_units_consistent(frame.props[1])


class TestCommon(object):
    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(default={
                "components": {"c1": {}, "c2": {}, "c3": {}},
                "phases": {
                    "a": {"equation_of_state": DummyEoS},
                    "b": {"equation_of_state": DummyEoS}},
                "state_definition": FTPx,
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "state_bounds": {
                    "flow_mol": (0, 0.1, 0.2, pyunits.kmol/pyunits.s),
                    "temperature": (522, 621, 720, pyunits.degR),
                    "pressure": (1, 3, 5, pyunits.bar)},
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K}})

        # Build state block
        m.props = m.params.build_state_block(
            [1], default={"defined_state": True})

        # Add necessary variables that would be built by other methods
        m.props[1].dens_mol_phase = Var(m.params.phase_list, initialize=1)
        m.props[1].enth_mol_phase = Var(m.params.phase_list, initialize=1)

        return m

    @pytest.mark.unit
    def test_convert_vars(self, frame):
        # Check that all state var values and bounds were converted correctly
        assert frame.props[1].flow_mol.value == 100
        assert frame.props[1].flow_mol.lb == 0
        assert frame.props[1].flow_mol.ub == 200
        assert check_units_equivalent(frame.props[1].flow_mol,
                                      pyunits.mol/pyunits.s)

        assert frame.props[1].pressure.value == 3e5
        assert frame.props[1].pressure.lb == 1e5
        assert frame.props[1].pressure.ub == 5e5
        assert check_units_equivalent(frame.props[1].pressure,
                                      pyunits.Pa)

        assert frame.props[1].temperature.value == 345
        assert frame.props[1].temperature.lb == 290
        assert frame.props[1].temperature.ub == 400
        assert check_units_equivalent(frame.props[1].temperature,
                                      pyunits.K)

        # Check supporting variables
        assert isinstance(frame.props[1].flow_mol_phase, Var)
        assert len(frame.props[1].flow_mol_phase) == 2

        assert isinstance(frame.props[1].mole_frac_phase_comp, Var)
        assert len(frame.props[1].mole_frac_phase_comp) == 6

        assert isinstance(frame.props[1].phase_frac, Var)
        assert len(frame.props[1].phase_frac) == 2

        assert isinstance(frame.props[1].total_flow_balance, Constraint)
        assert len(frame.props[1].total_flow_balance) == 1

        assert isinstance(frame.props[1].component_flow_balances, Constraint)
        assert len(frame.props[1].component_flow_balances) == 3

        assert isinstance(frame.props[1].sum_mole_frac, Constraint)
        assert len(frame.props[1].sum_mole_frac) == 1

        assert not hasattr(frame.props[1], "sum_mole_frac_out")

        assert isinstance(frame.props[1].phase_fraction_constraint, Constraint)
        assert len(frame.props[1].phase_fraction_constraint) == 2

        assert_units_consistent(frame)

    @pytest.mark.unit
    def test_calculate_scaling_factors(self, frame):
        frame.props[1].calculate_scaling_factors()

        assert len(frame.props[1].scaling_factor) == 22
        assert frame.props[1].scaling_factor[frame.props[1].flow_mol] == 1e-2
        assert frame.props[1].scaling_factor[
            frame.props[1].flow_mol_phase["a"]] == 1e-2
        assert frame.props[1].scaling_factor[
            frame.props[1].flow_mol_phase["b"]] == 1e-2
        assert frame.props[1].scaling_factor[
            frame.props[1].flow_mol_phase_comp["a", "c1"]] == 1e-2
        assert frame.props[1].scaling_factor[
            frame.props[1].flow_mol_phase_comp["a", "c2"]] == 1e-2
        assert frame.props[1].scaling_factor[
            frame.props[1].flow_mol_phase_comp["a", "c3"]] == 1e-2
        assert frame.props[1].scaling_factor[
            frame.props[1].flow_mol_phase_comp["b", "c1"]] == 1e-2
        assert frame.props[1].scaling_factor[
            frame.props[1].flow_mol_phase_comp["b", "c2"]] == 1e-2
        assert frame.props[1].scaling_factor[
            frame.props[1].flow_mol_phase_comp["b", "c3"]] == 1e-2
        assert frame.props[1].scaling_factor[
            frame.props[1].dens_mol_phase["a"]] == 1e-2
        assert frame.props[1].scaling_factor[
            frame.props[1].dens_mol_phase["b"]] == 1e-2
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_comp["c1"]] == 1e3
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_comp["c2"]] == 1e3
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_comp["c3"]] == 1e3
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_phase_comp["a", "c1"]] == 1e3
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_phase_comp["a", "c2"]] == 1e3
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_phase_comp["a", "c3"]] == 1e3
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_phase_comp["b", "c1"]] == 1e3
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_phase_comp["b", "c2"]] == 1e3
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_phase_comp["b", "c3"]] == 1e3
        assert frame.props[1].scaling_factor[frame.props[1].pressure] == 1e-5
        assert frame.props[1].scaling_factor[
            frame.props[1].temperature] == 1e-2

    # Test General Methods
    @pytest.mark.unit
    def test_get_material_flow_terms(self, frame):
        for (p, j) in frame.params._phase_component_set:
            assert str(frame.props[1].get_material_flow_terms(p, j)) == str(
                frame.props[1].flow_mol_phase_comp[p, j])

    @pytest.mark.unit
    def test_get_enthalpy_flow_terms(self, frame):
        for p in frame.params.phase_list:
            assert str(frame.props[1].get_enthalpy_flow_terms(p)) == str(
                frame.props[1]._enthalpy_flow_term[p])
            assert str(frame.props[1]._enthalpy_flow_term[p].expr) == str(
                frame.props[1].flow_mol_phase[p] *
                frame.props[1].enth_mol_phase[p])

    @pytest.mark.unit
    def test_get_material_density_terms(self, frame):
        for (p, j) in frame.params._phase_component_set:
            assert str(frame.props[1].get_material_density_terms(p, j)) == str(
                frame.props[1]._material_density_term[p, j])
            assert str(
                frame.props[1]._material_density_term[p, j].expr) == str(
                frame.props[1].dens_mol_phase[p] *
                frame.props[1].mole_frac_phase_comp[p, j])

    @pytest.mark.unit
    def test_get_energy_density_terms(self, frame):
        for p in frame.params.phase_list:
            assert str(frame.props[1].get_energy_density_terms(p)) == str(
                frame.props[1]._energy_density_term[p])
            assert str(
                frame.props[1]._energy_density_term[p].expr) == str(
                frame.props[1].dens_mol_phase[p] *
                frame.props[1].energy_internal_mol_phase[p])

    @pytest.mark.unit
    def test_default_material_balance_type(self, frame):
        assert frame.props[1].default_material_balance_type() == \
            MaterialBalanceType.componentTotal

    @pytest.mark.unit
    def test_default_energy_balance_type(self, frame):
        assert frame.props[1].default_energy_balance_type() == \
            EnergyBalanceType.enthalpyTotal

    @pytest.mark.unit
    def test_get_material_flow_basis(self, frame):
        assert frame.props[1].get_material_flow_basis() == \
            MaterialFlowBasis.molar

    @pytest.mark.unit
    def test_define_state_vars(self, frame):
        assert frame.props[1].define_state_vars() == \
            {"flow_mol": frame.props[1].flow_mol,
             "mole_frac_comp": frame.props[1].mole_frac_comp,
             "temperature": frame.props[1].temperature,
             "pressure": frame.props[1].pressure}

    @pytest.mark.unit
    def test_define_display_vars(self, frame):
        assert frame.props[1].define_display_vars() == \
            {"Total Molar Flowrate": frame.props[1].flow_mol,
             "Total Mole Fraction": frame.props[1].mole_frac_comp,
             "Temperature": frame.props[1].temperature,
             "Pressure": frame.props[1].pressure}

    @pytest.mark.unit
    def test_conc_mol(self, frame):
        assert isinstance(frame.props[1].conc_mol_comp, Expression)
        assert len(frame.props[1].conc_mol_comp) == 3
        assert isinstance(frame.props[1].conc_mol_phase_comp, Expression)
        assert len(frame.props[1].conc_mol_phase_comp) == 6



#!/usr/bin/env python
"""
Extract JSON schemas into files.
"""
import argparse
import json
import os
import sys
#
from idaes.dmf import resource

DMF_SOURCE_DIR = os.path.join('source', 'dmf')


def main():
    p = argparse.ArgumentParser()
    p.add_argument('-d', '--dir', dest='dir', default='schemas',
                   help='Destination directory '
                        'for .json files. If absolute, used as-is. '
                        'If relative, added to "{}". (default=schemas)'
                   .format(DMF_SOURCE_DIR))
    args = p.parse_args()
    if os.path.isabs(args.dir):
        schema_dir = args.dir
    else:
        schema_dir = os.path.join(DMF_SOURCE_DIR, args.dir)
    for schema, name in ((resource.RESOURCE_SCHEMA, 'resource'),):
        json_file = os.path.join(schema_dir, name) + '.json'
        json_fp = open(json_file, 'w')
        json.dump(schema, json_fp, indent=4)
        print('Wrote {}'.format(json_file))
    return 0


if __name__ == '__main__':
    sys.exit(main())



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
First Generation (GEN 1) Liquid Property Package.

This property package provides the necessary constraints or expressions for the
liquid phase properties of amine-based (MEA) scrubbing of CO2 acid gas
The GEN 1 MEA model uses the Enhancement factor calculation. MEA is taken to
be non-volatile.

Liquid phase components: Carbondioxide (CO2), Monoethanolamine (MEA), Water (H2O)

"""

# Import Pyomo libraries
from pyomo.environ import (Constraint, Expression, Param, SolverFactory,
                           Reference, PositiveReals, Reals, NonNegativeReals,
                           Set, value, Var, exp, log, units as pyunits)
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.common.config import ConfigValue, In


# Import IDAES libraries
from idaes.core import (declare_process_block_class,
                        MaterialFlowBasis,
                        PhysicalParameterBlock,
                        StateBlockData,
                        StateBlock,
                        Component,
                        LiquidPhase)
from idaes.core.util.initialization import (fix_state_vars,
                                            revert_state_vars,
                                            solve_indexed_blocks)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.power_generation.carbon_capture.mea_solvent_system.unit_models.column \
    import ProcessType
from idaes.core.util import get_solver
import idaes.logger as idaeslog

__author__ = "Paul Akula, John Eslick"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("LiquidParameterBlock")
class PhysicalParameterData(PhysicalParameterBlock):
    """
    Liquid Phase Property Parameter Block Class

    Contains parameters and indexing sets associated with
    liquid phase properties for amine-based scrubbing process.

    """
    CONFIG = PhysicalParameterBlock.CONFIG()
    CONFIG.declare("process_type", ConfigValue(
        default=ProcessType.absorber,
        domain=In(ProcessType),
        description="Flag indicating the type of  process",
        doc="""Flag indicating either absorption or stripping process.
            **default** - ProcessType.absorber.
            **Valid values:** {
            **ProcessType.absorber** - absorption process,
            **ProcessType.stripper** - stripping process.}"""))


    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(PhysicalParameterData, self).build()

        self._state_block_class = LiquidStateBlock

        # Create Phase object
        self.Liq = LiquidPhase()

        # Create Component objects
        self.MEA = Component()
        self.CO2 = Component()
        self.H2O = Component()

        # component list for true species
        self.component_list_true = Set(
            initialize=['CO2', 'H2O', 'MEA', 'MEA+', 'MEACOO-', 'HCO3-'])
        # component list for solvent species
        self.component_list_solvent = Set(initialize=['H2O', 'MEA'])
        # list of all diffusing components aside the excess solvent
        self.component_list_diffus = Set(
            initialize=['CO2', 'MEA', 'MEA+', 'MEACOO-'])
        # list of diffusing component into/from the vapor phase
        self.component_list_d = Set(initialize=['CO2', 'H2O'])

        # Reaction index
        self.reaction_idx = Set(initialize=['R1', 'R2'])

        # Thermodynamic reference state
        self.pressure_ref = Param(within=PositiveReals,
                                  mutable=True,
                                  default=101325,
                                  units=pyunits.Pa,
                                  doc='Reference pressure')
        self.temperature_ref = Param(default=298.15,
                                     units=pyunits.K,
                                     doc='Thermodynamic Reference Temperature')

        # Critical Temperature dictionary, unit K (Aprion 2005)
        self.temperature_crit = Param(
            self.component_list,
            mutable=False,
            initialize={'CO2': 304.18, 'H2O': 647.13, 'MEA': 614.45},
            units=pyunits.K,
            doc="Critical Temperature")

        # Mol. weights of vapor component - units = kg/mol.
        self.mw_comp = Param(
            self.component_list,
            mutable=False,
            initialize={'CO2': 0.04401, 'H2O': 0.01802, 'MEA': 0.06108},
            units=pyunits.kg / pyunits.mol,
            doc="Molecular weights of liquid components")

        '''
        Specific Heat capacity of Liquid,kJ/kgK
        Data from Hilliard thesis(1998)

        Cp (kJ/kgC) =  C1 + C2*t + C3*t^2 + C4*t^3 +C5*t^4
        t(C) = T(K) - a, a = 273.15
        C1 = C1
        C2*t = C2*(T-a)
        C3*t^2 = C3*(T^2 -2aT +a^2)
        C4*t^3 = C4*(T^3 -3aT^2 + 3a^2T -a^3)
        C5*t^4 = C5*(T^4 -4aT^3 + 6a^2T^2 -4a^3T + a^4)

        CpkJ/kgK = K1 + K2*T + K3*T^2 + K4*T^3 +K5*T^4
        required for mean cp
        K1 = C1 -aC2 + a^2C3 - a^3C4 + a^4C5
        K2 = C2 -2aC3 + 3a^2C4 -4a^3C5
        K3 = C3 -3aC4 + 6a^2C5
        K4 = C4 -4aC5
        K5 = C5
        '''
        _T = 273.15
        mea_c1 = 2.6161
        mea_c2 = 3.706e-3
        mea_c3 = 3.787e-6
        mea_c4 = 0.0
        mea_c5 = 0.0
        h2o_c1 = 4.2107
        h2o_c2 = -1.696e-3
        h2o_c3 = 2.568e-5
        h2o_c4 = -1.095e-7
        h2o_c5 = 3.038e-10

        mea_K1 = mea_c1 - _T * mea_c2 + _T**2 * \
            mea_c3 - _T**3 * mea_c4 + _T**4 * mea_c5
        mea_K2 = mea_c2 - 2 * _T * mea_c3 + 3 * _T**2 * mea_c4 - 4 * _T**3 * mea_c5
        mea_K3 = mea_c3 - 3 * _T * mea_c4 + 6 * _T**2 * mea_c5
        mea_K4 = mea_c4 - 4 * _T * mea_c5
        mea_K5 = mea_c5
        h2o_K1 = h2o_c1 - _T * h2o_c2 + _T**2 * \
            h2o_c3 - _T**3 * h2o_c4 + _T**4 * h2o_c5
        h2o_K2 = h2o_c2 - 2 * _T * h2o_c3 + 3 * _T**2 * h2o_c4 - 4 * _T**3 * h2o_c5
        h2o_K3 = h2o_c3 - 3 * _T * h2o_c4 + 6 * _T**2 * h2o_c5
        h2o_K4 = h2o_c4 - 4 * _T * h2o_c5
        h2o_K5 = h2o_c5

        cp_param_dict = {
            ('MEA', 1): mea_K1,
            ('MEA', 2): mea_K2,
            ('MEA', 3): mea_K3,
            ('MEA', 4): mea_K4,
            ('MEA', 5): mea_K5,
            ('H2O', 1): h2o_K1,
            ('H2O', 2): h2o_K2,
            ('H2O', 3): h2o_K3,
            ('H2O', 4): h2o_K4,
            ('H2O', 5): h2o_K5
        }
        self.cp_param = Param(self.component_list_solvent,
                              range(1, 6),
                              mutable=False,
                              initialize=cp_param_dict,
                              units=pyunits.kJ / (pyunits.kg * pyunits.K),
                              doc="Heat capacity parameters for solvent")

        # Viscosity constants for Aqueous MEA
        # Reference: Morgan et.al (2015)
        visc_d_param_dict = {
            1: -0.0838,
            2: 2.8817,
            3: 33.651,
            4: 1817.0,
            5: 0.00847,
            6: 0.0103,
            7: -2.3890
        }
        self.visc_d_param = Param(range(1, 8),
                                  mutable=False,
                                  initialize=visc_d_param_dict,
                                  units=pyunits.Pa * pyunits.s,
                                  doc="Dynamic viscosity constants")

        # Pure solvent molar volume parameters (MEA and H2O),
        # parameter unit in ml/mol or cm^3/mol
        # Reference: Morgan et.al (2015)
        vol_mol_comp_param_dict = {
            ('MEA', 1): -5.35162e-7,
            ('MEA', 2): -4.51417e-4,
            ('MEA', 3): 1.19451,
            ('H2O', 1): -3.2484e-6,
            ('H2O', 2): 0.00165,
            ('H2O', 3): 0.793,
        }
        self.vol_mol_comp_param = Param(self.component_list_solvent,
                                        range(1, 4),
                                        mutable=False,
                                        initialize=vol_mol_comp_param_dict,
                                        units=(pyunits.cm**3) / pyunits.mol,
                                        doc="Pure solvent molar volume constants")

        # Liquid molar volume parameters  Aqueous MEA,
        # parameter unit in ml/mol
        # Reference: Morgan et.al (2015)
        vol_mol_param_dict = {
            1: 10.2074,
            2: -2.2642,
            3: 3.0059,
            4: 207,
            5: -563.3701
        }
        self.vol_mol_param = Param(range(1, 6),
                                   mutable=False,
                                   initialize=vol_mol_param_dict,
                                   units=(pyunits.cm**3) / pyunits.mol,
                                   doc="Liquid molar volume constants")

        # Thermal conductivity parameters for water
        # k_H2O = C1 + C2*T/T_ref + C3(T/T_ref)^2, T_ref =298.15
        # DIPPR
        therm_cond_param_dict = {1: -0.9003,
                                 2: 2.5006,
                                 3: -0.9938
                                 }
        self.therm_cond_param = Param(range(1, 4),
                                      mutable=False,
                                      initialize=therm_cond_param_dict,
                                      units=pyunits.W / pyunits.m / pyunits.K,
                                      doc="Thermal conductivity constants")

        # Diffusion Coefficient(binary) constants, m^2/s
        # CO2 ---> Ying and Eimer (2012)
        # MEA ---> Snijder et al. (1993)
        # MEA+(MEAH+) ----> Hoff et al.(2004)
        # MEACOO-(MEACOO-)----> Hoff et al.(2004)
        diffus_param_dict = {
            ('CO2', 1): 2.35e-6,
            ('CO2', 2): 2.9837e-8,
            ('CO2', 3): -9.7078e-9,
            ('CO2', 4): -2119,
            ('CO2', 5): -20.132,
            ('MEA', 1): -13.275,
            ('MEA', 2): -2198.3,
            ('MEA', 3): -7.8142e-5,
            ('MEA', 4): 0.0,
            ('MEA', 5): 0.0,
            ('MEA+', 1): -22.64,
            ('MEA+', 2): -1000.0,
            ('MEA+', 3): -0.7,
            ('MEA+', 4): 0.0,
            ('MEA+', 5): 0.0,
            ('MEACOO-', 1): -22.64,
            ('MEACOO-', 2): -1000.0,
            ('MEACOO-', 3): -0.7,
            ('MEACOO-', 4): 0.0,
            ('MEACOO-', 5): 0.0
        }
        self.diffus_param = \
            Param(self.component_list_diffus, range(1, 6),
                  initialize=diffus_param_dict,
                  units=pyunits.m**2 / pyunits.s,
                  doc=" Diffusivity parameters")

        # Concentration-based Equilibruim constants -
        # Reference: Morgan et al 2017
        k_eq_conc_param_dict = {('R1', 1): 233.4,
                                ('R1', 2): -3410,
                                ('R1', 3): -36.8,
                                ('R2', 1): 176.72,
                                ('R2', 2): -2909,
                                ('R2', 3): -28.46
                                }
        self. k_eq_conc_param = Param(self.reaction_idx,
                                      range(1, 4),
                                      mutable=False,
                                      initialize=k_eq_conc_param_dict,
                                      units=pyunits.m**3 / pyunits.kmol,
                                      doc="Equilibrium constant parameters")

        # Surface tension constants, N/m
        # CO2 ---> Morgan et al. (2015)
        # MEA & H2O ---> Aprion(2005)
        # surface tension F parameter --->Morgan et al. (2015)
        surf_tens_param_dict = {
            ('CO2', 1): -5.987,
            ('CO2', 2): 3.7699,
            ('CO2', 3): -0.43164,
            ('CO2', 4): 0.018155,
            ('CO2', 5): -0.01207,
            ('CO2', 6): 0.002119,
            ('MEA', 1): 0.09945,
            ('MEA', 2): 1.067,
            ('MEA', 3): 0.0,
            ('MEA', 4): 0.0,
            ('MEA', 5): 0.0,
            ('MEA', 6): 0.0,
            ('H2O', 1): 0.18548,
            ('H2O', 2): 2.717,
            ('H2O', 3): -3.554,
            ('H2O', 4): 2.047,
            ('H2O', 5): 0.0,
            ('H2O', 6): 0.0
        }
        self.surf_tens_param = \
            Param(self.component_list, range(1, 7),
                  initialize=surf_tens_param_dict,
                  units=pyunits.N / pyunits.m,
                  doc=" Surface tension parameters")

        surf_tens_F_param_dict = {
            1: 2.4558,
            2: -1.5311,
            3: 3.4994,
            4: -5.6398,
            5: 10.2109,
            6: 2.3122,
            7: 4.5608,
            8: -2.3924,
            9: 5.3324,
            10: -12.0494
        }
        self.surf_tens_F_param = \
            Param(range(1, 11),
                  initialize=surf_tens_F_param_dict,
                  doc="Surface tension F parameters")

        # saturated vapor pressure parameters
        pressure_sat_param_dict = {
            ('MEA', 1): 172.78,
            ('MEA', 2): -13492,
            ('MEA', 3): -21.914,
            ('MEA', 4): 1.38e-5,
            ('H2O', 1): 72.55,
            ('H2O', 2): -7206.70,
            ('H2O', 3): -7.1385,
            ('H2O', 4): 4.05e-6
        }
        self.pressure_sat_param = \
            Param(self.component_list_solvent, range(1, 5),
                  initialize=pressure_sat_param_dict,
                  units=pyunits.Pa,
                  doc="Vapor pressure  parameters")

        # Heat of vaporization parameters of water
        heat_vaporization_param_dict = {
            1: 56.6,
            2: 0.61204,
            3: -0.6267,
            4: 0.3988}

        self.heat_vaporization_param = \
            Param(range(1, 5),
                  initialize=heat_vaporization_param_dict,
                  units=pyunits.kJ / pyunits.mol,
                  doc="Heat of vaporization parameters of water")

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({
            'flow_mol':
            {'method': None, 'units': 'mol/s'},
            'pressure':
            {'method': None, 'units': 'Pa'},
            'temperature':
            {'method': None, 'units': 'K'},
            'mole_frac_comp':
            {'method': None, 'units': None},
            'flow_mol_comp':
            {'method': '_flow_mol_comp', 'units': 'mol/s'},
            'mw':
            {'method': '_mw', 'units': 'kg/mol'},
            'mass_frac_comp':
            {'method': '_mass_frac_comp', 'units': None},
            'mass_frac_co2_free':
            {'method': '_mass_frac_co2_free', 'units': None},
            'vol_mol':
            {'method': '_vol_mol', 'units': 'm^3/mol'},
            'vol_mol_comp':
            {'method': '_vol_mol_comp', 'units': 'cm^3/mol'},
            'conc_mol':
            {'method': '_conc_mol', 'units': 'mol/m^3'},
            'conc_mol_comp':
            {'method': '_conc_mol_comp', 'units': 'mol/m^3'},
            'lnk_eq_conc':
            {'method': '_lnk_eq_conc', 'units': 'm^3/mol'},
            'dens_mass':
            {'method': '_dens_mass', 'units': 'kg/m^3'},
            'cp_mass_comp':
            {'method': '_cp_mass_comp', 'units': 'J/kg.K'},
            'cp_mol':
            {'method': '_cp_mol', 'units': 'J/mol.K'},
            'cp_mol_mean':
            {'method': '_cp_mol_mean', 'units': 'J/mol.K'},
            'cp_mol_comp':
            {'method': '_cp_mol_comp', 'units': 'J/mol.K'},
            'cp_mol_comp_mean':
            {'method': '_cp_mol_comp_mean', 'units': 'J/mol.K'},
            'cp_mass_comp_mean':
            {'method': '_cp_mass_comp_mean', 'units': 'J/kg.K'},
            'enth_mean':
            {'method': '_enth_mean', 'units': 'J/s'},
            'enth_liq_density':
            {'method': '_enth_liq_density', 'units': 'J/m^3'},
            'diffus':
            {'method': '_diffus', 'units': 'm^2/s'},
            'visc_d':
            {'method': '_visc_d', 'units': 'kg/m.s'},
            'thermal_cond':
            {'method': '_thermal_cond', 'units': 'W/m.K'},
            'surf_tens':
            {'method': '_surf_tens', 'units': 'N/m'},
            'pressure_sat':
            {'method': '_pressure_sat', 'units': 'Pa'},
            'k2_rxn':
            {'method': '_k2_rxn', 'units': 'm3/mol.s'},
            'habs':
            {'method': '_habs', 'units': 'J/mol'},
            'hvap':
            {'method': '_hvap', 'units': 'J/mol'},
            'henry_N2O_analogy':
            {'method': '_henry_N2O_analogy', 'units': 'Pa.m3/mol'}
        })

        obj.add_default_units({'time': pyunits.s,
                               'length': pyunits.m,
                               'mass': pyunits.kg,
                               'amount': pyunits.mol,
                               'temperature': pyunits.K})


class LiquidStateBlockMethods(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """
    def initialize(blk, state_args=None,
                   state_vars_fixed=False,
                   hold_state=False, outlvl=idaeslog.NOTSET,
                   solver=None, optarg=None):
        """
        Initialization routine for property package.

        Keyword Arguments:
          state_args : Dictionary with initial guesses for the state vars
                       chosen. Note that if this method is triggered through
                       the control volume, and if initial guesses were not
                       provided at the unit model level, the control volume
                       passes the inlet values as initial guess.Keys for the
                       state_args dictionary are: flow_mol, temperature,
                       pressure and mole_frac_comp.
          outlvl : sets output level of initialization routine
          optarg : solver options dictionary object (default=None, use
                     default solver options)
          solver : str indicating which solver to use during
                   initialization (default = None)
          hold_state :
                  flag indicating whether the initialization routine
                  should unfix any state variables fixed during initialization
                  (default=False).

                  valid options:
                    True :
                      states varaibles are not unfixed, and a dict of returned
                      containing flags for which states were fixed during
                      initialization.
                    False :
                      state variables are unfixed after initialization by
                      calling the relase_state method
        Returns:
          If hold_states is True, returns a dict containing flags for which
          states were fixed during initialization.

        """

        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="properties")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="properties")

        init_log.info('Starting Liquid phase properties initialization')

        # Deactivate the constraints specific for non-inlet blocks i.e.
        # when defined state is False
        for k in blk.keys():
            if blk[k].config.defined_state is False:
                blk[k].sum_component_eqn.deactivate()

        # Fix state variables if not already fixed
        if state_vars_fixed is False:
            flags = fix_state_vars(blk, state_args)

        for k in blk.keys():
            dof = degrees_of_freedom(blk[k])
            if dof != 0:
                raise RuntimeError(
                    "{} - degrees of freedom for state block is not zero "
                    "during initialization. DoF = {}".format(blk.name, dof))

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialise values
        for k in blk.keys():
            for j in blk[k].params.component_list_solvent:
                if hasattr(blk[k], "cp_mass_comp_eqn"):
                    calculate_variable_from_constraint(blk[k].cp_mass_comp[j],
                                                       blk[k].cp_mass_comp_eqn[j])
                if hasattr(blk[k], "cp_mol_comp_eqn"):
                    calculate_variable_from_constraint(blk[k].cp_mol_comp[j],
                                                       blk[k].cp_mol_comp_eqn[j])

                if hasattr(blk[k], "flow_mol_comp_eqn"):
                    calculate_variable_from_constraint(blk[k].flow_mol_comp[j],
                                                       blk[k].flow_mol_comp_eqn[j])

                if hasattr(blk[k], "cp_mass_comp_mean_eqn"):
                    calculate_variable_from_constraint(
                        blk[k].cp_mass_comp_mean[j],
                        blk[k].cp_mass_comp_mean_eqn[j])

                if hasattr(blk[k], "cp_mol_comp_mean_eqn"):
                    calculate_variable_from_constraint(
                        blk[k].cp_mol_comp_mean[j],
                        blk[k].cp_mol_comp_mean_eqn[j])

            if hasattr(blk[k], "cp_mol_eqn"):
                calculate_variable_from_constraint(blk[k].cp_mol,
                                                   blk[k].cp_mol_eqn)

            if hasattr(blk[k], "cp_mol_mean_eqn"):
                calculate_variable_from_constraint(
                    blk[k].cp_mol_mean,
                    blk[k].cp_mol_mean_eqn)

            if hasattr(blk[k], "enth_mean_eqn"):
                calculate_variable_from_constraint(
                    blk[k].enth_mean,
                    blk[k].enth_mean_eqn)

        # Initialise values of true species
        for k in blk.keys():
            if hasattr(blk[k], "speciation_model"):
                blk[k].conc_mol_comp_true['CO2'].value = 0.01539
                blk[k].conc_mol_comp_true['MEA'].value = 2032
                blk[k].conc_mol_comp_true['H2O'].value = 3.209e+04
                blk[k].conc_mol_comp_true['MEA+'].value = 2007
                blk[k].conc_mol_comp_true['HCO3-'].value = 24.43
                blk[k].conc_mol_comp_true['MEACOO-'].value = 1983

        # Solve property block if non-empty
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solve_indexed_blocks(opt, [blk], tee=slc.tee)

        init_log.info("Liquid properties initialization complete {}.".format(
            idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                blk.release_state(flags)

    def release_state(blk, flags, outlvl=idaeslog.NOTSET):
        """
        Method to release state variables fixed during initialisation.

        Keyword Arguments:
            flags : dict containing information of which state variables were fixed
                    during initialization, and should now be unfixed. This dict is
                    returned by initialize if hold_state=True.
            outlvl : sets output level of of logging
        """

        if flags is None:
            return

        # Unfix state variables
        revert_state_vars(blk, flags)

        # Activate state variable related constraints
        for k in blk.keys():
            if blk[k].config.defined_state is False:
                blk[k].sum_component_eqn.activate()

        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="properties")
        init_log.info_high('States released.')


@declare_process_block_class("LiquidStateBlock",
                             block_class=LiquidStateBlockMethods)
class LiquidStateBlockData(StateBlockData):
    """
    Liquid phase property package of amine-based scrubbing of CO2
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(LiquidStateBlockData, self).build()

        # Object reference for molecular weight if needed by CV1D
        # Molecular weights
        self.mw_comp = Reference(self.params.mw_comp)

        self._make_state_vars()
        self._make_speciation_model()

    def _make_speciation_model(self):
        self.conc_mol_comp_true = Var(self.params.component_list_true,
                                      domain=NonNegativeReals,
                                      initialize=10,
                                      units=pyunits.mol / pyunits.m**3,
                                      doc="Concentration of true species ")

        def rule_logC(blk, i):
            return log(blk.conc_mol_comp_true[i])

        self.logC = Expression(self.params.component_list_true,
                               rule=rule_logC,
                               doc='Logarithm of true species concentration')

        def rule_speciation_model(blk, r):
            xa = blk.conc_mol_comp
            xt = blk.conc_mol_comp_true
            CT = blk.logC
            lnk = blk.lnk_eq_conc

            # chemical equilibruim reaction 1
            if r == 1:
                return lnk['R1'] == \
                    CT['MEA+'] + CT['MEACOO-'] - 2 * CT['MEA'] - CT['CO2']
            # chemical equilibruim reaction 2
            elif r == 2:
                return lnk['R2'] == \
                    CT['MEA+'] + CT['HCO3-'] - \
                    CT['MEA'] - CT['CO2'] - CT['H2O']
            # MEA balance
            elif r == 3:
                return xa['MEA'] == xt['MEA'] + xt['MEA+'] + xt['MEACOO-']
            # CO2 balance
            elif r == 4:
                return xa['CO2'] == xt['CO2'] + xt['HCO3-'] + xt['MEACOO-']
            # H2O balance
            elif r == 5:
                return xa['H2O'] == xt['H2O'] + xt['HCO3-']
            # charge balance
            elif r == 6:
                return xt['MEA+'] == xt['HCO3-'] + xt['MEACOO-']

        self.speciation_model = Constraint([1, 2, 3, 4, 5, 6],
                                           rule=rule_speciation_model)

    def _make_state_vars(self):

        self.flow_mol = Var(initialize=1.0,
                            domain=NonNegativeReals,
                            units=pyunits.mol / pyunits.s,
                            doc='Total molar flowrate')
        self.mole_frac_comp = Var(self.component_list,
                                  domain=NonNegativeReals,
                                  bounds=(0, 1),
                                  initialize=1 /
                                  len(self.component_list),
                                  units=None,
                                  doc='Component mole fractions')
        self.pressure = Var(initialize=101325,
                            domain=NonNegativeReals,
                            units=pyunits.Pa,
                            doc='Pressure [Pa]')
        self.temperature = Var(initialize=298.15,
                               domain=NonNegativeReals,
                               units=pyunits.K,
                               doc='Temperature')

        # Sum mole fractions if not inlet block
        if self.config.defined_state is False:
            def sum_component_eqn(b):
                return  b.flow_mol == sum(b.flow_mol_comp[j]
                                        for j in b._params.component_list)
            self.sum_component_eqn = Constraint(rule=sum_component_eqn)

    def _flow_mol_comp(self):

        def rule_flow_mol_comp(b, j):
            return b.flow_mol_comp[j] == b.mole_frac_comp[j] * b.flow_mol

        try:
            self.flow_mol_comp = Var(self.component_list,
                                     initialize=1.0,
                                     domain=NonNegativeReals,
                                     doc='Component molar flowrate [mol/s]')
            self.flow_mol_comp_eq = Constraint(self.component_list,
                                               rule=rule_flow_mol_comp,
                                               doc="Component molar flow in"
                                                   "liquid phase [mol/s]")
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.flow_mol_comp)
            self.del_component(self.flow_mol_comp_eq)
            raise

    def _mw(self):
        def rule_mw(blk):
            return sum(blk.mole_frac_comp[j] * blk.mw_comp[j]
                       for j in blk.component_list)

        try:
            self.mw = Expression(rule=rule_mw,
                                 doc="Average molecular weight of  liquid phase"
                                 "components")
        except AttributeError:
            self.del_component(self.mw)
            raise

    def _mass_frac_comp(self):
        def rule_mass_frac_comp(blk, j):
            return blk.mole_frac_comp[j] * blk.mw_comp[j] / blk.mw

        try:
            self.mass_frac_comp = Expression(self.component_list,
                                             rule=rule_mass_frac_comp,
                                             doc="Mass fraction"
                                             "[-]")
        except AttributeError:
            self.del_component(self.mass_frac_comp)
            raise

    def _mass_frac_co2_free(self):
        '''
        mass fraction on CO2 free-basis
        '''
        def rule_mass_frac_co2_free(blk, j):
            return blk.mole_frac_comp[j] * blk.mw_comp[j] /\
                sum(blk.mole_frac_comp[i] * blk.mw_comp[i]
                    for i in blk.params.component_list_solvent)

        try:
            self.mass_frac_co2_free = Expression(self.params.component_list_solvent,
                                                 rule=rule_mass_frac_co2_free,
                                                 doc="Mass fraction on CO2 free basis"
                                                 "[-]")
        except AttributeError:
            self.del_component(self.mass_frac_co2_free)
            raise

    def _vol_mol_comp(self):
        # UNIT: cm**3/mol,
        def rule_vol_mol_comp(blk, j):
            T = blk.temperature
            mw = blk.mw_comp
            c = blk._params.vol_mol_comp_param
            return mw[j] * 1e3 / (c[j, 1] * T**2 + c[j, 2] * T + c[j, 3])

        try:
            self.vol_mol_comp = Expression(self.params.component_list_solvent,
                                           rule=rule_vol_mol_comp,
                                           doc="Pure solvent molar volume")
        except AttributeError:
            self.del_component(self.vol_mol_comp)
            raise

    def _vol_mol(self):
        # Liquid phase molar volume m^3/mol
        def rule_vol_mol(blk):
            v_mea = self.vol_mol_comp['MEA']
            v_h20 = self.vol_mol_comp['H2O']
            a = blk._params.vol_mol_param[1]
            b = blk._params.vol_mol_param[2]
            c = blk._params.vol_mol_param[3]
            d = blk._params.vol_mol_param[4]
            e = blk._params.vol_mol_param[5]
            v_soln = (blk.mole_frac_comp['MEA'] * v_mea + blk.mole_frac_comp['H2O']
                      * v_h20 + a * blk.mole_frac_comp['CO2'] +
                      (b + c * blk.mole_frac_comp['MEA'])
                      * blk.mole_frac_comp['MEA'] * blk.mole_frac_comp['H2O'] +
                      (d + e * blk.mole_frac_comp['MEA']) *
                      blk.mole_frac_comp['MEA'] * blk.mole_frac_comp['CO2']) * 1e-6
            return v_soln

        try:
            self.vol_mol = Expression(rule=rule_vol_mol,
                                      doc="Liquid molar volume")
        except AttributeError:
            self.del_component(self.vol_mol)
            raise

    def _conc_mol(self):
        # Liquid phase concentration
        def rule_conc_mol(b):
            return 1 / self.vol_mol

        try:
            self.conc_mol = Expression(rule=rule_conc_mol,
                                       doc="Concentration [mol/m3]")
        except AttributeError:
            self.del_component(self.conc_mol)
            raise

    def _conc_mol_comp(self):
        def rule_conc_mol_comp(blk, i):
            return blk.conc_mol * blk.mole_frac_comp[i]

        try:
            self.conc_mol_comp = Expression(self.component_list,
                                            rule=rule_conc_mol_comp,
                                            doc="Concentration of liquid components")
        except AttributeError:
            self.del_component(self.conc_mol_comp)
            raise

    def _lnk_eq_conc(self):
        '''
        Natural logarithm of concentration based equilibruim constant [m**3/mol]
        for chemical equilibruim in the liquid phase.
        '''
        def rule_lnk_eq_conc(blk, rxn):
            return (self._params.k_eq_conc_param[rxn, 1] +
                    self._params.k_eq_conc_param[rxn, 2] / blk.temperature +
                    self._params.k_eq_conc_param[rxn, 3] * log(blk.temperature) +
                    log(1e-3))

        try:
            self.lnk_eq_conc = Expression(self._params.reaction_idx,
                                          rule=rule_lnk_eq_conc,
                                          doc="Natural log of concentration"
                                          "based equilibruim constant [m3/mol]")
        except AttributeError:
            self.del_component(self.lnk_eq_conc)
            raise

    def _dens_mass(self):
        def rule_dens_mass(blk):
            return blk.mw * blk.conc_mol

        try:
            self.dens_mass = Expression(rule=rule_dens_mass,
                                        doc="mass density [kg/m^3]")
        except AttributeError:
            self.del_component(self.dens_mass)
            raise

    def _cp_mass_comp(self):
        self.cp_mass_comp = Var(self.params.component_list_solvent,
                                domain=Reals,
                                initialize=1.0,
                                units=pyunits.J / pyunits.K / pyunits.kg,
                                doc="Pure component specific heat capacities ")

        def rule_cp_mass_comp(blk, j):
            return blk.cp_mass_comp[j] == 1e3 * (
                blk._params.cp_param[j, 1] +
                blk._params.cp_param[j, 2] * (blk.temperature) +
                blk._params.cp_param[j, 3] * (blk.temperature)**2 +
                blk._params.cp_param[j, 4] * (blk.temperature)**3 +
                blk._params.cp_param[j, 5] * (blk.temperature)**4)

        try:
            # Try to build constraint
            self.cp_mass_comp_eqn = Constraint(self.params.component_list_solvent,
                                               rule=rule_cp_mass_comp)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.cp_mass_comp)
            self.del_component(self.cp_mass_comp_eqn)
            raise

    def _cp_mol_comp(self):
        # Pure component liquid heat capacities J/mol.K
        self.cp_mol_comp = Var(self.params.component_list_solvent,
                               domain=Reals,
                               initialize=1.0,
                               units=pyunits.J / pyunits.K / pyunits.mol,
                               doc="Pure component liquid heat capacities ")

        def rule_cp_mol_comp(blk, j):
            return blk.cp_mol_comp[j] == blk.cp_mass_comp[j] * blk.mw_comp[j]

        try:
            # Try to build constraint
            self.cp_mol_comp_eqn = Constraint(self.params.component_list_solvent,
                                              rule=rule_cp_mol_comp)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.cp_mol_comp)
            self.del_component(self.cp_mol_comp_eqn)
            raise

    def _cp_mol(self):
        self.cp_mol = Var(domain=Reals,
                          initialize=1.0,
                          units=pyunits.J / pyunits.K / pyunits.mol,
                          doc="Liquid molar heat capacity ")

        def rule_cp_mol(blk):
            return blk.cp_mol == sum(blk.cp_mol_comp[j] * blk.mole_frac_comp[j]
                                     for j in blk.params.component_list_solvent)

        try:
            # Try to build constraint
            self.cp_mol_eqn = Constraint(rule=rule_cp_mol)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.cp_mol)
            self.del_component(self.cp_mol_eqn)
            raise

    def _cp_mass_comp_mean(self):
        # mean Pure component liquid heat capacities
        self.cp_mass_comp_mean = Var(self.params.component_list_solvent,
                                     domain=Reals,
                                     initialize=1.0,
                                     units=pyunits.J / pyunits.K / pyunits.kg,
                                     doc="Mean Pure component specific heat capacities "
                                         "between T and T_ref ")

        def rule_cp_mass_comp_mean(blk, j):
            t = blk.temperature / blk._params.temperature_ref
            t0 = blk._params.temperature_ref
            return blk.cp_mass_comp_mean[j] == 1e3 * (
                blk._params.cp_param[j, 1] +
                blk._params.cp_param[j, 2] / 2 * t0 * (t + 1) +
                blk._params.cp_param[j, 3] / 3 * (t0**2) * (t**2 + t + 1) +
                blk._params.cp_param[j, 4] / 4 * (t0**3) * (t + 1) * (t**2 + 1) +
                blk._params.cp_param[j, 5] / 5 * (t0**4) * (t**4 + t**3 + t**2 + t + 1))

        try:
            # Try to build constraint
            self.cp_mass_comp_mean_eqn =\
                Constraint(self.params.component_list_solvent,
                           rule=rule_cp_mass_comp_mean)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.cp_mass_comp_mean)
            self.del_component(self.cp_mass_comp_mean_eqn)
            raise

    def _cp_mol_comp_mean(self):
        # average Pure component liquid heat capacities btw T and T_ref
        self.cp_mol_comp_mean = Var(self.params.component_list_solvent,
                                    domain=Reals,
                                    initialize=1.0,
                                    units=pyunits.J / pyunits.K / pyunits.mol,
                                    doc="Mean pure component liq. heat capacities "
                                        "between T and T_ref ")

        def rule_cp_mol_comp_mean(blk, j):
            return blk.cp_mol_comp_mean[j] == (blk.cp_mass_comp_mean[j] *
                                               blk.mw_comp[j])

        try:
            # Try to build constraint
            self.cp_mol_comp_mean__eqn = Constraint(self.params.component_list_solvent,
                                                    rule=rule_cp_mol_comp_mean)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.cp_mol_comp_mean)
            self.del_component(self.cp_mol_comp_mean_eqn)
            raise

    def _cp_mol_mean(self):
        # Average liquid heat capacities btw T and T_ref
        self.cp_mol_mean = Var(domain=Reals,
                               initialize=1.0,
                               units=pyunits.J / pyunits.K / pyunits.mol,
                               doc="Mean molar liquid heat capacities ")

        def rule_cp_mol_mean(blk):
            return blk.cp_mol_mean == sum(blk.cp_mol_comp_mean[j] *
                                          blk.mole_frac_comp[j]
                                          for j in blk.params.component_list_solvent)

        try:
            # Try to build constraint
            self.cp_mol_mean_eqn = Constraint(rule=rule_cp_mol_mean)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.cp_mol_mean)
            self.del_component(self.cp_mol_mean_eqn)
            raise

    def _enth_mean(self):
        # Average Liquid Phase Enthalpy  btw T and T_ref
        self.enth_mean = Var(domain=Reals,
                                 initialize=1.0,
                                 units=pyunits.J / pyunits.s,
                                 doc="Mean Liquid Enthalpy flow  btw T and Tref ")

        def rule_enth_mean(b):
            return b.enth_mean == b.cp_mol_mean * b.flow_mol * b.temperature

        try:
            # Try to build constraint
            self.enth_mean_eqn = Constraint(rule=rule_enth_mean)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.enth_mean)
            self.del_component(self.enth_mean_eqn)
            raise

    def _enth_liq_density(self):
        #  molar enthalpy holdup per unit volume
        self.enth_liq_density = Var(
            domain=Reals,
            initialize=1.0,
            units=pyunits.J / pyunits.m**3,
            doc=' Enthalpy holdup')
        try:
            # Try to build constraint
            self.eq_enth_liq_density = Constraint(
                expr=self.enth_liq_density == self.enth_mean * self.conc_mol)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.enth_liq_density)
            self.del_component(self.eq_enth_liq_density)
            raise

    def _visc_d(self):
        '''
        Dynamic viscosity of liquid
        '''
        def rule_visc_d(blk):
            r = blk.mass_frac_comp['MEA'] * 100
            T = blk.temperature
            alpha = blk.mole_frac_comp['CO2'] / blk.mole_frac_comp['MEA']
            mu_H2O = 1.002e-3 *\
                10**((1.3272 *
                      (293.15 - T - 0.001053 * (T - 293.15)**2)) / (T - 168.15))
            a = blk._params.visc_d_param[1]
            b = blk._params.visc_d_param[2]
            c = blk._params.visc_d_param[3]
            d = blk._params.visc_d_param[4]
            e = blk._params.visc_d_param[5]
            f = blk._params.visc_d_param[6]
            g = blk._params.visc_d_param[7]

            return mu_H2O * exp(r * (T * (a * r + b) + c * r + d) *
                                (alpha * (e * r + f * T + g) + 1) / T**2)

        try:
            self.visc_d = Expression(rule=rule_visc_d,
                                     doc="Dynamic viscosity [Pa.s]")

        except AttributeError:
            self.del_component(self.visc_d)
            raise

    def _thermal_cond(self):
        '''
        thermal conductivity of fluid[W/m/K]
        source:
        pure components :Sato-Riedel model
        mixing rule: Vredeveld mixing rule
        arguments
        T = temperature of the liquid phase
        Tb_MEA= boiling point of MEA [K]
        Tc_MEA = critical Temperature of MEA [K]
        Tr = reduced Temperature of the liquid
        Tbr_MEA = reduced boiling point temperature of MEA
        K_MEA = thermal conductivity of pure MEA
        K_H2O = thermal conductivity of pure water
        '''
        def rule_thermal_cond(blk):
            Tb_MEA = 443
            Tc_MEA = 614.2
            T = blk.temperature
            Tr = T / Tc_MEA
            Tbr_MEA = Tb_MEA / Tc_MEA
            K_MEA = 1.1053152 / (61.08**0.5) * (3 + 20 * (1 - Tr)
                                                ** (2 / 3)) / (3 + 20 *
                                                               (1 - Tbr_MEA)**(2 / 3))
            K_H2O = 0.6065 * (-1.48445 + 4.12292 * T / 298.15 -
                              1.63866 * (T / 298.15)**2)

            return ((blk.mole_frac_comp['H2O'] * K_H2O**-2 +
                     blk.mole_frac_comp['MEA'] * K_MEA**-2)**-1)**0.5

        try:
            self.thermal_cond = Expression(rule=rule_thermal_cond,
                                           doc='Thermal conductivity of fluid [W/m.K]')
        except AttributeError:
            self.del_component(self.thermal_cond)
            raise

    def _surf_tens(self):
        '''
        Surface tension,N/m
        '''
        def rule_surf_tens_comp(blk, j):
            r = blk.mass_frac_co2_free['MEA']
            T = blk.temperature
            Tc = blk._params.temperature_crit
            if j in blk.params.component_list_solvent:
                return blk._params.surf_tens_param[j, 1] * (1 - T / Tc[j]) **\
                    (blk._params.surf_tens_param[j, 2] +
                     blk._params.surf_tens_param[j, 3] * T / Tc[j] +
                     blk._params.surf_tens_param[j, 4] * (T / Tc[j])**2)
            elif j == 'CO2':
                return blk._params.surf_tens_param[j, 1] * r**2 + \
                    blk._params.surf_tens_param[j, 2] * r + \
                    blk._params.surf_tens_param[j, 3] +\
                    T * (blk._params.surf_tens_param[j, 4] * r**2 +
                         blk._params.surf_tens_param[j, 5] * r +
                         blk._params.surf_tens_param[j, 6])

        try:
            self.surf_tens_comp = Expression(self.component_list,
                                             rule=rule_surf_tens_comp,
                                             doc="Surface tension of components")
        except AttributeError:
            self.del_component(self.surf_tens_comp)
            raise

        def rule_surf_tens(blk):
            r = blk.mass_frac_co2_free['MEA']
            alpha = self.mole_frac_comp['CO2'] / blk.mole_frac_comp['MEA']
            Fa = blk._params.surf_tens_F_param[1]
            Fb = blk._params.surf_tens_F_param[2]
            Fc = blk._params.surf_tens_F_param[3]
            Fd = blk._params.surf_tens_F_param[4]
            Fe = blk._params.surf_tens_F_param[5]
            Ff = blk._params.surf_tens_F_param[6]
            Fg = blk._params.surf_tens_F_param[7]
            Fh = blk._params.surf_tens_F_param[8]
            Fi = blk._params.surf_tens_F_param[9]
            Fj = blk._params.surf_tens_F_param[10]
            return blk.surf_tens_comp['H2O'] + \
                (blk.surf_tens_comp['CO2'] - blk.surf_tens_comp['H2O']) *\
                blk.mole_frac_comp['CO2'] * (Fa + Fb * alpha + Fc * alpha**2 +
                                             Fd * r + Fe * r**2) +\
                (blk.surf_tens_comp['MEA'] - blk.surf_tens_comp['H2O']) *\
                (Ff + Fg * alpha + Fh * alpha**2 + Fi * r + Fj * r**2) *\
                blk.mole_frac_comp['MEA']

        try:
            self.surf_tens = Expression(rule=rule_surf_tens,
                                        doc="Surface tension [N/m] ")
        except AttributeError:
            self.del_component(self.surf_tens)
            raise

    def _pressure_sat(self):
        '''
        saturated vapor pressure of solvents
        '''
        def rule_pressure_sat_comp(blk, j):
            T = blk.temperature
            return (exp(blk._params.pressure_sat_param[j, 1] +
                        blk._params.pressure_sat_param[j, 2] / T +
                        blk._params.pressure_sat_param[j, 3] * log(T) +
                        blk._params.pressure_sat_param[j, 4] * T * T))

        try:
            self.pressure_sat = Expression(self.params.component_list_solvent,
                                           rule=rule_pressure_sat_comp,
                                           doc="Vapor pressure [Pa]")
        except AttributeError:
            self.del_component(self.pressure_sat)
            raise

    def _diffus(self):
        """
        Diffusivity of diffusing components, m**2/s
        """
        def rule_diffus(blk, i):
            if i == 'CO2':
                C_MEA = blk.conc_mol_comp['MEA'] * 1e-3
                return (blk._params.diffus_param[i, 1] +
                        blk._params.diffus_param[i, 2] * C_MEA +
                        blk._params.diffus_param[i, 3] * C_MEA**2) * exp(
                    (blk._params.diffus_param[i, 4] +
                     blk._params.diffus_param[i, 5]) / blk.temperature)
            elif i == 'MEA':
                C_MEA = blk.conc_mol_comp['MEA']
                return exp(blk._params.diffus_param[i, 1] +
                           blk._params.diffus_param[i, 2] / blk.temperature +
                           blk._params.diffus_param[i, 3] * C_MEA)
            elif i == "MEACOO-" or i == 'MEA+':
                return exp(blk._params.diffus_param[i, 1] +
                           blk._params.diffus_param[i, 2] / blk.temperature +
                           blk._params.diffus_param[i, 3] * log(blk.visc_d))

        try:
            self.diffus = Expression(self.params.component_list_diffus,
                                     rule=rule_diffus,
                                     doc='Diffusivity of component i in liquid')
        except AttributeError:
            self.del_component(self.diffus)
            raise

    def _k2_rxn(self):
        '''
        concentration based rate constant for overall reaction rate
        Second order rate contant(Luo et al., 2015)
        '''
        def rule_k2_rxn(blk):
            T = blk.temperature
            C_MEA = blk.conc_mol_comp['MEA'] * 1e-3
            C_H2O = blk.conc_mol_comp['H2O'] * 1e-3
            return (2.003e10 * exp(-4742 / T) * C_MEA +
                    4.147e6 * exp(-3110 / T) * C_H2O) * 1e-3

        try:
            self.k2_rxn = Expression(rule=rule_k2_rxn,
                                     doc="Second order rate contant [m3/(mol.s)]")
        except AttributeError:
            self.del_component(self.k2_rxn)
            raise

    def _habs(self):
        '''
        Heat of absorption of CO2 J/mol
        '''
        def rule_habs(blk):
            if blk.config.parameters.config.process_type == ProcessType.stripper:
                return -97000
            elif blk.config.parameters.config.process_type == ProcessType.absorber:
                return -84000
        try:
            self.habs = Expression(rule=rule_habs,
                                   doc="CO2 heat of absorption")
        except AttributeError:
            self.del_component(self.habs)
            raise

    def _hvap(self):
        '''
        Heat of vaporization of H2O
        '''
        def rule_hvap(blk):
            c = blk._params.heat_vaporization_param
            T = blk.temperature
            Tc = blk._params.temperature_crit['H2O']
            return 1000 * c[1] * (1 - T / Tc)**(c[2] + c[3] *
                                                T / Tc + c[4] * (T / Tc)**2)
        try:
            self.hvap = Expression(rule=rule_hvap,
                                   doc="Heat of vaporization of H2O [J/mol]")
        except AttributeError:
            self.del_component(self.hvap)
            raise

    def _henry_N2O_analogy(self):
        '''
        Henry's constant N2O Analogy Jiru et.al (2012)
        '''
        def rule_henry_N2O_analogy(blk):
            T = blk.temperature
            t = blk.temperature - 273.15
            wt_MEA = blk.mass_frac_co2_free['MEA']
            wt_H2O = blk.mass_frac_co2_free['H2O']
            H_N2O_MEA = 2.448e5 * exp(-1348 / T)
            H_CO2_H2O = 3.52e6 * exp(-2113 / T)
            H_N2O_H2O = 8.449e6 * exp(-2283 / T)
            H_CO2_MEA = H_N2O_MEA * (H_CO2_H2O / H_N2O_H2O)
            lwm = 1.70981 + 0.03972 * t - 4.3e-4 * t**2 - 2.20377 * wt_H2O

            return (exp(wt_MEA * log(H_CO2_MEA) + wt_H2O * log(H_CO2_H2O) +
                        wt_MEA * wt_H2O * lwm))

        try:
            self.henry_N2O_analogy = Expression(
                rule=rule_henry_N2O_analogy,
                doc="Henry's constant [Pa.m3/mol]")
        except AttributeError:
            self.del_component(self.henry_N2O_analogy)
            raise
    # --------------------------------------------------------------------------

    def get_material_flow_terms(self, p, j):
        return self.flow_mol_comp[j]

    def get_enthalpy_flow_terms(self, p):
        return self.enth_mean

    def get_material_density_terms(self, p, j):
        return self.conc_mol_comp[j]

    def get_energy_density_terms(self, p):
        return self.enth_liq_density

    def define_state_vars(self):
        return {"flow_mol": self.flow_mol,
                "temperature": self.temperature,
                "pressure": self.pressure,
                "mole_frac_comp": self.mole_frac_comp}

    def get_material_flow_basis(self):
        return MaterialFlowBasis.molar

    def model_check(blk):
        """
        Model checks for property block
        """
        # Check temperature bounds
        if value(blk.temperature) < blk.temperature.lb:
            _log.error('{} Temperature set below lower bound.'
                       .format(blk.name))
        if value(blk.temperature) > blk.temperature.ub:
            _log.error('{} Temperature set above upper bound.'
                       .format(blk.name))

        # Check pressure bounds
        if value(blk.pressure) < blk.pressure.lb:
            _log.error('{} Pressure set below lower bound.'.format(blk.name))
        if value(blk.pressure) > blk.pressure.ub:
            _log.error('{} Pressure set above upper bound.'.format(blk.name))



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Standard IDAES CSTR model.
"""

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.environ import Reference, Block, Var, Constraint

# Import IDAES cores
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        UnitModelBlockData,
                        useDefault)
from idaes.core.util.config import (is_physical_parameter_block,
                                    is_reaction_parameter_block)
import idaes.core.util.unit_costing as costing

__author__ = "Andrew Lee, Vibhav Dabadghao"


@declare_process_block_class("CSTR")
class CSTRData(UnitModelBlockData):
    """
    Standard CSTR Unit Model Class
    """
    CONFIG = UnitModelBlockData.CONFIG()

    CONFIG.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.useDefault,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.useDefault,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.useDefault.
**Valid values:** {
**EnergyBalanceType.useDefault - refer to property package for default
balance type
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))
    CONFIG.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))
    CONFIG.declare("has_heat_transfer", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Heat transfer term construction flag",
        doc="""Indicates whether terms for heat transfer should be constructed,
**default** - False.
**Valid values:** {
**True** - include heat transfer terms,
**False** - exclude heat transfer terms.}"""))
    CONFIG.declare("has_pressure_change", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Pressure change term construction flag",
        doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}"""))
    CONFIG.declare("has_equilibrium_reactions", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Equilibrium reaction construction flag",
        doc="""Indicates whether terms for equilibrium controlled reactions
should be constructed,
**default** - True.
**Valid values:** {
**True** - include equilibrium reaction terms,
**False** - exclude equilibrium reaction terms.}"""))
    CONFIG.declare("has_phase_equilibrium", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Phase equilibrium construction flag",
        doc="""Indicates whether terms for phase equilibrium should be
constructed,
**default** = False.
**Valid values:** {
**True** - include phase equilibrium terms
**False** - exclude phase equilibrium terms.}"""))
    CONFIG.declare("has_heat_of_reaction", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Heat of reaction term construction flag",
        doc="""Indicates whether terms for heat of reaction terms should be
constructed,
**default** - False.
**Valid values:** {
**True** - include heat of reaction terms,
**False** - exclude heat of reaction terms.}"""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))
    CONFIG.declare("reaction_package", ConfigValue(
        default=None,
        domain=is_reaction_parameter_block,
        description="Reaction package to use for control volume",
        doc="""Reaction parameter object used to define reaction calculations,
**default** - None.
**Valid values:** {
**None** - no reaction package,
**ReactionParameterBlock** - a ReactionParameterBlock object.}"""))
    CONFIG.declare("reaction_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing reaction packages",
        doc="""A ConfigBlock with arguments to be passed to a reaction block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see reaction package for documentation.}"""))

    def build(self):
        """
        Begin building model (pre-DAE transformation).
        Args:
            None
        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(CSTRData, self).build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(default={
                "dynamic": self.config.dynamic,
                "has_holdup": self.config.has_holdup,
                "property_package": self.config.property_package,
                "property_package_args": self.config.property_package_args,
                "reaction_package": self.config.reaction_package,
                "reaction_package_args": self.config.reaction_package_args})

        self.control_volume.add_geometry()

        self.control_volume.add_state_blocks(
                has_phase_equilibrium=self.config.has_phase_equilibrium)

        self.control_volume.add_reaction_blocks(
                has_equilibrium=self.config.has_equilibrium_reactions)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_rate_reactions=True,
            has_equilibrium_reactions=self.config.has_equilibrium_reactions,
            has_phase_equilibrium=self.config.has_phase_equilibrium)

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_of_reaction=self.config.has_heat_of_reaction,
            has_heat_transfer=self.config.has_heat_transfer)

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change)

        # Add Ports
        self.add_inlet_port()
        self.add_outlet_port()

        # Add object references
        self.volume = Reference(self.control_volume.volume[:])

        # Add CSTR performance equation
        @self.Constraint(self.flowsheet().config.time,
                         self.config.reaction_package.rate_reaction_idx,
                         doc="CSTR performance equation")
        def cstr_performance_eqn(b, t, r):
            return b.control_volume.rate_reaction_extent[t, r] == (
                            b.volume[t] *
                            b.control_volume.reactions[t].reaction_rate[r])

        # Set references to balance terms at unit level
        if (self.config.has_heat_transfer is True and
                self.config.energy_balance_type != EnergyBalanceType.none):
            self.heat_duty = Reference(self.control_volume.heat[:])

        if (self.config.has_pressure_change is True and
                self.config.momentum_balance_type != MomentumBalanceType.none):
            self.deltaP = Reference(self.control_volume.deltaP[:])

    def _get_performance_contents(self, time_point=0):
        var_dict = {"Volume": self.volume[time_point]}
        if hasattr(self, "heat_duty"):
            var_dict["Heat Duty"] = self.heat_duty[time_point]
        if hasattr(self, "deltaP"):
            var_dict["Pressure Change"] = self.deltaP[time_point]

        return {"vars": var_dict}

    def get_costing(self, year=None, module=costing, **kwargs):
        if not hasattr(self.flowsheet(), "costing"):
            self.flowsheet().get_costing(year=year, module=module)

        self.costing = Block()
        units_meta = (self.config.property_package.get_metadata().
                      get_derived_units)
        self.length = Var(initialize=1,
                          units=units_meta('length'),
                          doc='vessel length')
        self.diameter = Var(initialize=1,
                            units=units_meta('length'),
                            doc='vessel diameter')
        time = self.flowsheet().config.time.first()
        self.volume_eq = Constraint(expr=self.volume[time]
                                    == self.length*self.diameter)
        module.cstr_costing(self.costing, **kwargs)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""IDAES IAPWS-95 Steam properties

Dropped all critical enhancments and non-analytic terms ment to improve
accruacy near the critical point. These tend to cause singularities in the
equations, and it is assumend that we will try to avoid operating very close to
the critical point.

 References: (some of this is only used in the C++ part)
   International Association for the Properties of Water and Steam (2016).
       IAPWS R6-95 (2016), "Revised Release on the IAPWS Formulation 1995 for
       the Properties of Ordinary Water Substance for General Scientific Use,"
       URL: http://iapws.org/relguide/IAPWS95-2016.pdf
   Wagner, W.,  A. Pruss (2002). "The IAPWS Formulation 1995 for the
       Thermodynamic Properties of Ordinary Water Substance for General and
       Scientific Use." J. Phys. Chem. Ref. Data, 31, 387-535.
   Wagner, W. et al. (2000). "The IAPWS Industrial Formulation 1997 for the
       Thermodynamic Properties of Water and Steam," ASME J. Eng. Gas Turbines
       and Power, 122, 150-182.
   Akasaka, R. (2008). "A Reliable and Useful Method to Determine the
       Saturation State from Helmholtz Energy Equations of State." Journal of
       Thermal Science and Technology, 3(3), 442-451.
   International Association for the Properties of Water and Steam (2011).
       IAPWS R15-11, "Release on the IAPWS Formulation 2011 for the
       Thermal Conductivity of Ordinary Water Substance,"
       URL: http://iapws.org/relguide/ThCond.pdf
   International Association for the Properties of Water and Steam (2008).
       IAPWS R12-08, "Release on the IAPWS Formulation 2008 for the Viscosity
       of Ordinary Water Substance,"
       URL: http://iapws.org/relguide/visc.pdf
"""
__author__ = "John Eslick"

# Import Python libraries
import os

# Import Pyomo libraries
from pyomo.environ import (
    Expression,
    Param,
    RangeSet,
    Set,
    exp,
    sqrt,
    units as pyunits
)

# Import IDAES
from idaes.core import declare_process_block_class

import idaes
import idaes.logger as idaeslog
from idaes.generic_models.properties.helmholtz.helmholtz import (
    _available,
    _htpx,
    HelmholtzParameterBlockData,
    HelmholtzStateBlockData,
    HelmholtzThermoExpressions,
    PhaseType,
    StateVars,
    _StateBlock,
)

# Logger
_log = idaeslog.getLogger(__name__)
_so = os.path.join(idaes.bin_directory, "iapws95_external.so")


def iapws95_available():
    """Make sure the compiled IAPWS-95 functions are available. Yes, in Windows
    the .so extention is still used.
    """
    return _available(_so)


def htpx(T=None, P=None, x=None):
    """
    Convenience function to calculate steam enthalpy from temperature and
    either pressure or vapor fraction. This function can be used for inlet
    streams and initialization where temperature is known instead of enthalpy.
    User must provided values for two of T, P, or x.

    Args:
        T: Temperature with units (between 200 and 3000 K)
        P: Pressure with units (between 1 and 1e9 Pa), None if saturated vapor
        x: Vapor fraction [mol vapor/mol total] (between 0 and 1), None if
        superheated or subcooled

    Returns:
        Total molar enthalpy [J/mol].
    """
    prop = Iapws95StateBlock(default={"parameters": Iapws95ParameterBlock()})
    return _htpx(T=T, P=P, x=x, prop=prop,
                 Tmin=270*pyunits.K, Tmax=3e3*pyunits.K,
                 Pmin=1e-4*pyunits.kPa, Pmax=1e6*pyunits.kPa)


@declare_process_block_class("Iapws95ParameterBlock")
class Iapws95ParameterBlockData(HelmholtzParameterBlockData):
    CONFIG = HelmholtzParameterBlockData.CONFIG()

    def build(self):
        self._set_parameters(
            library=_so,
            eos_tag="iapws95",
            state_block_class=Iapws95StateBlock,
            component_list=Set(initialize=["H2O"]),
            phase_equilibrium_idx=Set(initialize=[1]),
            phase_equilibrium_list={1: ["H2O", ("Vap", "Liq")]},
            mw=Param(initialize=0.01801528,
                     doc="Molecular weight [kg/mol]",
                     units=pyunits.kg/pyunits.mol),
            temperature_crit=Param(
                initialize=647.096,
                doc="Critical temperature [K]",
                units=pyunits.K),
            pressure_crit=Param(initialize=2.2064e7,
                                doc="Critical pressure [Pa]",
                                units=pyunits.Pa),
            dens_mass_crit=Param(initialize=322,
                                 doc="Critical density [kg/m3]",
                                 units=pyunits.kg/pyunits.m**3),
            specific_gas_constant=Param(
                initialize=461.51805,
                doc="Water Specific Gas Constant [J/kg/K]",
                units=pyunits.J/pyunits.kg/pyunits.K),
            pressure_bounds=(0.1, 1e9),
            temperature_bounds=(250, 2500),
            enthalpy_bounds=(0, 1e5),
        )
        super().build()
        # Thermal conductivity parameters.
        # "Release on the IAPWS Formulation 2011 for the Thermal Conductivity
        # of Ordinary Water Substance"
        self.tc_L0 = Param(
            RangeSet(0, 5),
            initialize={
                0: 2.443221e0,
                1: 1.323095e1,
                2: 6.770357e0,
                3: -3.454586e0,
                4: 4.096266e-1,
            },
            doc="0th order thermal conductivity parameters",
            units=pyunits.K*pyunits.m/pyunits.W
        )

        self.tc_L1 = Param(
            RangeSet(0, 5),
            RangeSet(0, 6),
            initialize={
                (0, 0): 1.60397357,
                (1, 0): 2.33771842,
                (2, 0): 2.19650529,
                (3, 0): -1.21051378,
                (4, 0): -2.7203370,
                (0, 1): -0.646013523,
                (1, 1): -2.78843778,
                (2, 1): -4.54580785,
                (3, 1): 1.60812989,
                (4, 1): 4.57586331,
                (0, 2): 0.111443906,
                (1, 2): 1.53616167,
                (2, 2): 3.55777244,
                (3, 2): -0.621178141,
                (4, 2): -3.18369245,
                (0, 3): 0.102997357,
                (1, 3): -0.463045512,
                (2, 3): -1.40944978,
                (3, 3): 0.0716373224,
                (4, 3): 1.1168348,
                (0, 4): -0.0504123634,
                (1, 4): 0.0832827019,
                (2, 4): 0.275418278,
                (3, 4): 0.0,
                (4, 4): -0.19268305,
                (0, 5): 0.00609859258,
                (1, 5): -0.00719201245,
                (2, 5): -0.0205938816,
                (3, 5): 0.0,
                (4, 5): 0.012913842,
            },
            doc="1st order thermal conductivity parameters",
        )
        # Viscosity parameters
        # "Release on the IAPWS Formulation 2008 for the Viscosity of
        # Ordinary Water Substance "
        self.visc_H0 = Param(
            RangeSet(0, 4),
            initialize={0: 1.67752, 1: 2.20462, 2: 0.6366564, 3: -0.241605},
            doc="0th order viscosity parameters",
            units=1/pyunits.s/pyunits.Pa
        )

        self.visc_H1 = Param(
            RangeSet(0, 6),
            RangeSet(0, 7),
            initialize={
                (0, 0): 5.20094e-1,
                (1, 0): 8.50895e-2,
                (2, 0): -1.08374,
                (3, 0): -2.89555e-1,
                (4, 0): 0.0,
                (5, 0): 0.0,
                (0, 1): 2.22531e-1,
                (1, 1): 9.99115e-1,
                (2, 1): 1.88797,
                (3, 1): 1.26613,
                (4, 1): 0.0,
                (5, 1): 1.20573e-1,
                (0, 2): -2.81378e-1,
                (1, 2): -9.06851e-1,
                (2, 2): -7.72479e-1,
                (3, 2): -4.89837e-1,
                (4, 2): -2.57040e-1,
                (5, 2): 0.0,
                (0, 3): 1.61913e-1,
                (1, 3): 2.57399e-1,
                (2, 3): 0.0,
                (3, 3): 0.0,
                (4, 3): 0.0,
                (5, 3): 0.0,
                (0, 4): -3.25372e-2,
                (1, 4): 0.0,
                (2, 4): 0.0,
                (3, 4): 6.98452e-2,
                (4, 4): 0.0,
                (5, 4): 0.0,
                (0, 5): 0.0,
                (1, 5): 0.0,
                (2, 5): 0.0,
                (3, 5): 0.0,
                (4, 5): 8.72102e-3,
                (5, 5): 0.0,
                (0, 6): 0.0,
                (1, 6): 0.0,
                (2, 6): 0.0,
                (3, 6): -4.35673e-3,
                (4, 6): 0.0,
                (5, 6): -5.93264e-4,
            },
            doc="1st order viscosity parameters",
        )
        self.set_default_scaling("therm_cond_phase", 1e2, index="Liq")
        self.set_default_scaling("therm_cond_phase", 1e1, index="Vap")
        self.set_default_scaling("visc_d_phase", 1e5, index="Liq")
        self.set_default_scaling("visc_d_phase", 1e6, index="Vap")
        self.set_default_scaling("visc_k_phase", 1e5, index="Liq")
        self.set_default_scaling("visc_k_phase", 1e7, index="Vap")


@declare_process_block_class(
    "Iapws95StateBlock",
    block_class=_StateBlock,
    doc="""This is some placeholder doc.""",
)
class Iapws95StateBlockData(HelmholtzStateBlockData):
    """
    This is a property package for calculating thermophysical properties of
    water.
    """

    def build(self, *args):
        """
        Callable method for Block construction
        """
        super().build(*args)

        phlist = self.config.parameters.private_phase_list
        tau = self.tau
        delta = self.dens_phase_red  # this shorter name is from IAPWS

        # Phase Thermal conductiviy
        def rule_tc(b, p):
            L0 = self.config.parameters.tc_L0
            L1 = self.config.parameters.tc_L1
            return (
                sqrt(1.0 / tau) / sum(L0[i] * tau ** i for i in L0) *
                exp(delta[p] * sum((tau - 1)**i *
                    sum(L1[i, j] * (delta[p] - 1)**j for j in range(0, 6))
                for i in range(0, 5))))

        self.therm_cond_phase = Expression(
            phlist, rule=rule_tc, doc="Thermal conductivity [W/K/m]"
        )

        # Phase dynamic viscosity
        def rule_mu(b, p):
            H0 = self.config.parameters.visc_H0
            H1 = self.config.parameters.visc_H1
            # The units of this are really weird, so I am just going to append
            # units to the expression rather than give units to the parameters
            return (
                1e-4 * sqrt(1.0 / tau) / sum(H0[i] * tau ** i for i in H0) *
                exp(delta[p] * sum((tau - 1)**i *
                    sum(H1[i, j] * (delta[p] - 1)**j for j in range(0, 7))
                for i in range(0, 6))))

        self.visc_d_phase = Expression(
            phlist, rule=rule_mu, doc="Viscosity (dynamic)"
        )

        # Phase kinimatic viscosity
        def rule_nu(b, p):
            return self.visc_d_phase[p] / self.dens_mass_phase[p]

        self.visc_k_phase = Expression(
            phlist, rule=rule_nu, doc="Kinematic viscosity"
        )



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, University of Notre Dame, et al. 
# All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This module contains model diagnostic utility functions for use in IDAES (Pyomo) models.
"""

import pytest

# Need to update
import pyomo.environ as pyo
                           
# TODO: Add pyomo.dae test case
'''
from pyomo.environ import TransformationFactory
from pyomo.dae import ContinuousSet, DerivativeVar
'''

# Need to update
from idaes.core.util.model_diagnostics import *

# Author: Alex Dowling

# This was from 
# @pytest.fixture()
def problem1():
    m = pyo.ConcreteModel()

    m.I = pyo.Set(initialize=[i for i in range(5)])

    m.x = pyo.Var(m.I,bounds=(-10,10),initialize=1.0)

    m.con1 = pyo.Constraint(expr=m.x[0] + m.x[1] - m.x[3] >= 10)
    m.con2 = pyo.Constraint(expr=m.x[0]*m.x[3] + m.x[1] >= 0)
    m.con3 = pyo.Constraint(expr=m.x[4]*m.x[3] + m.x[0]*m.x[3] - m.x[4] == 0)

    m.obj = pyo.Objective(expr=sum(m.x[i]**2 for i in m.I))

    return m

def example2(with_degenerate_constraint=True):
    ''' Create the Pyomo model for Example 2
    
    Arguments:
        with_degenerate_constraint: Boolean, if True, include the redundant linear constraint
    
    Returns:
        m2: Pyomo model
    '''
    
    m2 = pyo.ConcreteModel()

    m2.I = pyo.Set(initialize=[i for i in range(1,4)])

    m2.x = pyo.Var(m2.I,bounds=(0,5),initialize=1.0)

    m2.con1 = pyo.Constraint(expr=m2.x[1] + m2.x[2] >= 1)
    m2.con2 = pyo.Constraint(expr=m2.x[1] + m2.x[2] + m2.x[3] == 1)
    m2.con3 = pyo.Constraint(expr=m2.x[2] - 2*m2.x[3] <= 1)
    m2.con4 = pyo.Constraint(expr=m2.x[1] + m2.x[3] >= 1)
    
    if with_degenerate_constraint:
        m2.con5 = pyo.Constraint(expr=m2.x[1] + m2.x[2] + m2.x[3] == 1)

    m2.obj = pyo.Objective(expr=sum(m2.x[i] for i in m2.I))
    
    return m2

def extract_constraint_names(cs):
    ''' Get constraint names from ComponentSet
    
    Arguments:
        cs: ComponentSet object
        
    Return:
        constraint_names: list of constraint names (strings)
            
    '''
    
    constraint_names = []
    for i in cs:
        constraint_names.append(i.name)
    return constraint_names

# Problem 1
@pytest.mark.skipif(not pyo.SolverFactory('ipopt').available(False), reason="no Ipopt")
@pytest.mark.unit
def test_problem1():
    # Create test problem
    m = problem1()
    
    # Specify Ipopt as the solver
    opt = pyo.SolverFactory('ipopt')

    # Specifying an iteration limit of 0 allows us to inspect the initial point
    opt.options['max_iter'] = 0

    # "Solving" the model with an iteration limit of 0 load the initial point and applies
    # any preprocessors (e.g., enforces bounds)
    opt.solve(m, tee=True)

    # Create Degeneracy Hunter object
    dh = DegeneracyHunter(m) 
    
    # Find constraints with residuals > 0.1 
    initial_point_constraints = dh.check_residuals(tol=0.1)
    
    # Check there are 2 constraints with large residuals
    assert len(initial_point_constraints) == 2
    
    initial_point_constraint_names = extract_constraint_names(initial_point_constraints)
    
    # Check first constraint
    assert initial_point_constraint_names[0] == 'con1'
    
    # Check second constraint
    assert initial_point_constraint_names[1] == 'con3'
    
    opt.options['max_iter'] = 50
    
    # Solve
    opt.solve(m, tee=True)
    
    # Find constraints with residuals > 0.1 
    solution_constraints = dh.check_residuals(tol=1E-6)
    
    # Check at the solution no constraints are violated
    assert len(solution_constraints) == 0
    
    # Check no constraints are near their bounds
    solution_bounds = dh.check_variable_bounds(tol=0.1)
    
    # Check at the solution no constraints are violated
    assert len(solution_bounds) == 0
    

# Problem 2 without degenerate constraint
@pytest.mark.skipif(not pyo.SolverFactory('ipopt').available(False), reason="no Ipopt")
@pytest.mark.unit
def test_problem2_without_degenerate_constraint():

    # Create test problem instance
    m2 = example2(with_degenerate_constraint=False)
    
    # Specify Ipopt as the solver
    opt = pyo.SolverFactory('ipopt')
    
    # Specifying an iteration limit of 0 allows us to inspect the initial point
    opt.options['max_iter'] = 0

    # "Solving" the model with an iteration limit of 0 load the initial point and applies
    # any preprocessors (e.g., enforces bounds)
    opt.solve(m2, tee=True)
    
    # Create Degeneracy Hunter object
    dh2 = DegeneracyHunter(m2)
    
    # Check for violated constraints at the initial point
    initial_point_constraints = dh2.check_residuals(tol=0.1)
    
    # Check there are 1 constraints with large residuals
    assert len(initial_point_constraints) == 1
    
    initial_point_constraint_names = extract_constraint_names(initial_point_constraints)
    
    # Check first constraint
    assert initial_point_constraint_names[0] == 'con2'
    
    # Resolve
    opt.options['max_iter'] = 500
    opt.solve(m2, tee=True)
    
    # Check solution
    x_sln = []
    
    for i in m2.I:
        x_sln.append(m2.x[i]())
    
    assert pytest.approx(x_sln[0], abs=1E-6) == 1.0
    assert pytest.approx(x_sln[1], abs=1E-6) == 0.0
    assert pytest.approx(x_sln[2], abs=1E-6) == 0.0
    



# Problem 2 with degenerate constraint
@pytest.mark.skipif(not pyo.SolverFactory('ipopt').available(False), reason="no Ipopt")
@pytest.mark.unit
def test_problem2_with_degenerate_constraint():

    # Create test problem instance
    m2 = example2(with_degenerate_constraint=True)
    
    # Specify Ipopt as the solver
    opt = pyo.SolverFactory('ipopt')

    # Specifying an iteration limit of 0 allows us to inspect the initial point
    opt.options['max_iter'] = 0

    # "Solving" the model with an iteration limit of 0 load the initial point and applies
    # any preprocessors (e.g., enforces bounds)
    opt.solve(m2, tee=True)
    
    # Create Degeneracy Hunter object
    dh2 = DegeneracyHunter(m2)
    
    # Check for violated constraints at the initial point
    initial_point_constraints = dh2.check_residuals(tol=0.1)
    
    # Check there are 2 constraints with large residuals
    assert len(initial_point_constraints) == 2
    
    initial_point_constraint_names = extract_constraint_names(initial_point_constraints)
    
    # Check first constraint
    assert initial_point_constraint_names[0] == 'con2'
    
    # Check first constraint
    assert initial_point_constraint_names[1] == 'con5'
    
    # Resolve
    opt.options['max_iter'] = 500
    opt.solve(m2, tee=True)
    
    # Check solution
    x_sln = []
    
    for i in m2.I:
        x_sln.append(m2.x[i]())
    
    assert pytest.approx(x_sln[0], abs=1E-6) == 1.0
    assert pytest.approx(x_sln[1], abs=1E-6) == 0.0
    assert pytest.approx(x_sln[2], abs=1E-6) == 0.0
    
    # Check the rank
    n_rank_deficient = dh2.check_rank_equality_constraints()
    
    assert n_rank_deficient == 1
    
    # TODO: Add MILP solver to idaes get-extensions and add more tests